/*
 * puzzle.c
 *
 * by Korbinian Strimmer and Arndt von Haeseler (1995-96)
 * this file belongs to the PUZZLE version 2.5.1 distribution
 *
 */

#define EXTERN
#define TRUE 1
#define FALSE 0

#define VERSION    "2.5.1"
#define DATE       "1 October 1996"

#define INFILE     "infile"
#define OUTFILE    "outfile"
#ifdef SIMULATION
#define TREEFILE   "treefile"
#else
#define TREEFILE   "outtree"
#endif
#define INTREE     "intree"
#define DISTANCES  "outdist"
#define TRIANGLE   "triangle"
#define UNRESOLVED "outqlist"


/* prototypes */
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <ctype.h>
#include "matrixut.h"
#include "ml.h"

/* abbreviation */
typedef unsigned long int ulint;

#ifndef CONTBAYES

/* tree structure */
typedef struct oneedge {
	/* pointer to other three edges */
	struct oneedge *up;
	struct oneedge *downleft;
	struct oneedge *downright;
	int numedge;      /* number of edge */
	ulint edgeinfo;   /* value of this edge */
	int *edgemap;     /* pointer to the local edgemap */
} ONEEDGE;

#else

/* tree structure */
typedef struct oneedge {
	/* pointer to other three edges */
	struct oneedge *up;
	struct oneedge *downleft;
	struct oneedge *downright;
	int numedge;      /* number of edge */
	double edgescore; /* score of edge for continous weights */
	int *edgemap;     /* pointer to the local edgemap */
} ONEEDGE;

#endif

/* variables */
cmatrix biparts;      /* bipartitions of tree of current puzzling step */
cmatrix consbiparts;  /* bipartitions of majority rule consensus tree */
cmatrix seqchars;     /* characters contained in data set */
cmatrix treepict;     /* picture of consensus tree */
double minscore;      /* value of edgescore on minedge */
double tstvf84;       /* F84 transition/transversion ratio */
double tstvratio;     /* expected transition/transversion ratio */
double yrtsratio;     /* expected pyrimidine/purine transition ratio */
FILE *itfp;           /* pointer to intree file */
FILE *ofp;            /* pointer to output file */
FILE *seqfp;          /* pointer to sequence input file */
FILE *tfp;            /* pointer to tree file */
FILE *dfp;            /* pointer to distance file */
FILE *trifp;          /* pointer to triangle file */
FILE *unresfp;        /* pointer to file with unresolved quartets */
int *edgeofleaf;      /* vector with edge number of all leaves */
int chooseA;          /* leaf variable */
int chooseB;          /* leaf variable */
int codelength;       /* how many characters define one state */
int column;           /* used for breaking lines (writing tree to treefile) */
int Frequ_optn;       /* use empirical base frequencies */
int Maxbrnch;         /* 2*Maxspc - 3 */
int Maxseqc;          /* number of sequence characters per taxum */
int minedge;          /* edge with minimum edgeinfo */
int nextedge;         /* number of edges in the current tree */
int nextleaf;         /* next leaf to add to tree */
int outgroup;         /* outgroup */
int puzzlemode;       /* computation of QP tree and/or ML distances */
int splitlength;      /* length of one entry in splitpatterns */
int *splitsizes;      /* size of all different splits of all trees */
int sym_optn;         /* symmetrize doublet frequencies */
int xsize;            /* depth of consensus tree picture */
int ytaxcounter;      /* counter for establishing y-coordinates of all taxa */
ivector consconfid;   /* confidence values of majority rule consensus tree */
ivector conssizes;    /* partition sizes of majority rule consensus tree */
ivector trueID;       /* leaf -> taxon on this leaf */
ivector xcor;         /* x-coordinates of consensus tree nodes */
ivector ycor;         /* y-coordinates of consensus tree nodes */
ivector ycormax;      /* maximal y-coordinates of consensus tree nodes */
ivector ycormin;      /* minimal y-coordinates of consensus tree nodes */
ivector ycortax;      /* y-coordinates of all taxa */
ONEEDGE *edge;        /* vector with all the edges of the tree */
time_t Starttime;     /* start time */
time_t Stoptime;      /* stop time */
time_t time0;         /* timer variable */
time_t time1;         /* yet another timer */
time_t time2;         /* yet another timer */
ulint *splitcomp;     /* bipartition storage */
ulint *splitfreqs;    /* frequencies of all different splits of all trees */
ulint *splitpatterns; /* all different splits of all trees */
ulint badqs;          /* number of bad quartets */
ulint consincluded;   /* number of included biparts in the consensus tree */
ulint Currtrial;      /* counter for puzzling steps */
ulint maxbiparts;     /* space is reserved for that many bipartitions */
ulint mininfo;        /* value of edgeinfo on minedge */
ulint numbiparts;     /* number of different bipartitions */
ulint Numquartets;    /* number of quartets */
ulint Numtrial;       /* number of puzzling steps */

#ifndef CONTBAYES

unsigned char *quartetinfo; /* place where quartets are stored */

#else

double *quartetw1;   /* place where quartet weights are stored (w1) */
double *quartetw2;   /* place where quartet weights are stored (w2) */
double *quartetw3;   /* place where quartet weights are stored (w3) */

#endif

/* prototypes */
int ulintcmp();       /* compare function for quicksort */
int intcmp();         /* compare function for quicksort */
int dcmp();           /* compare function for quicksort */
int fputid(FILE *ofp, int t);   /* print taxon name */
void fputid10(FILE *ofp, int t);


/******************************************************************************/
/* random numbers generator  (Numerical recipes)                              */
/******************************************************************************/

/* definitions */
#define IM1 2147483563
#define IM2 2147483399
#define AM (1.0/IM1)
#define IMM1 (IM1-1)
#define IA1 40014
#define IA2 40692
#define IQ1 53668
#define IQ2 52774
#define IR1 12211
#define IR2 3791
#define NTAB 32
#define NDIV (1+IMM1/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)

/* variable */
long idum;

double randomunitintervall()
/* Long period (> 2e18) random number generator. Returns a uniform random
   deviate between 0.0 and 1.0 (exclusive of endpoint values).

   Source:
   Press et al., "Numerical recipes in C", Cambridge University Press, 1992
   (chapter 7 "Random numbers", ran2 random number generator) */
{
	int j;
	long k;
	static long idum2=123456789;
	static long iy=0;
	static long iv[NTAB];
	double temp;

	if (idum <= 0) {
		if (-(idum) < 1)
			idum=1;
		else
			idum=-(idum);
		idum2=(idum);
		for (j=NTAB+7;j>=0;j--) {
			k=(idum)/IQ1;
			idum=IA1*(idum-k*IQ1)-k*IR1;
			if (idum < 0)
				idum += IM1;
			if (j < NTAB)
				iv[j] = idum;
		}
		iy=iv[0];
	}
	k=(idum)/IQ1;
	idum=IA1*(idum-k*IQ1)-k*IR1;
	if (idum < 0)
		idum += IM1;
	k=idum2/IQ2;
	idum2=IA2*(idum2-k*IQ2)-k*IR2;
	if (idum2 < 0)
		idum2 += IM2;
	j=iy/NDIV;
	iy=iv[j]-idum2;
	iv[j] = idum;
	if (iy < 1)
		iy += IMM1;
	if ((temp=AM*iy) > RNMX)
		return RNMX;
	else
		return temp;
}

#undef IM1
#undef IM2
#undef AM
#undef IMM1
#undef IA1
#undef IA2
#undef IQ1
#undef IQ2
#undef IR1
#undef IR2
#undef NTAB
#undef NDIV
#undef EPS
#undef RNMX

void initrandom()
{
	srand((unsigned) time(NULL));
	idum=-(long) rand();
}

/* returns a random integer in the range [0; n - 1] */
int randominteger(int n)
{
	return (int) floor(randomunitintervall()*n);
}


/******************************************************************************/
/* functions for representing and building trees                              */
/******************************************************************************/

/* initialize tree with the following starting configuration

                 2
          0  +------- C(=2)
  A(=0) -----+
             +------- B(=1)
                 1
                                */
void inittree()
{
	int i;

	/* allocate the memory for the whole tree */
	
	/* allocate memory for vector with all the edges of the tree */
	edge = (ONEEDGE *) calloc(Maxbrnch, sizeof(ONEEDGE) );
	if (edge == NULL) maerror("edge [inittree]");

	/* allocate memory for vector with edge numbers of leaves */ 
	edgeofleaf = (int *) calloc(Maxspc, sizeof(int) );
	if (edgeofleaf == NULL) maerror("edgeofleaf [inittree]");
	
	/* allocate memory for all the edges the edge map */
	for (i = 0; i < Maxbrnch; i++) {
		edge[i].edgemap = (int *) calloc(Maxbrnch, sizeof(int) );
		if (edge[i].edgemap == NULL) maerror("edgemap [inittree]");
	}
		
	/* number all edges */
	for (i = 0; i < Maxbrnch; i++) edge[i].numedge = i;
	
	/* initialize tree */
	
	nextedge = 3;
	nextleaf = 3;
	
	/* edge maps */
	(edge[0].edgemap)[0] = 0; /* you are on the right edge */
	(edge[0].edgemap)[1] = 4; /* go down left for leaf 1 */
	(edge[0].edgemap)[2] = 5; /* go down right for leaf 2 */
	(edge[1].edgemap)[0] = 1; /* go up for leaf 0 */
	(edge[1].edgemap)[1] = 0; /* you are on the right edge */
	(edge[1].edgemap)[2] = 3; /* go up/down right for leaf 2 */
	(edge[2].edgemap)[0] = 1; /* go up for leaf 0 */
	(edge[2].edgemap)[1] = 2; /* go up/down left for leaf 1 */
	(edge[2].edgemap)[2] = 0; /* you are on the right edge */
	
	/* interconnection */
	edge[0].up = NULL;
	edge[0].downleft = &edge[1];
	edge[0].downright = &edge[2];
	edge[1].up = &edge[0];
	edge[1].downleft = NULL;
	edge[1].downright = NULL;
	edge[2].up = &edge[0];
	edge[2].downleft = NULL;
	edge[2].downright = NULL;
	
	/* edges of leaves */
	edgeofleaf[0] = 0;
	edgeofleaf[1] = 1;
	edgeofleaf[2] = 2;
}

/* add next leaf on the specified edge */
void addnextleaf(int dockedge)
{	
	int i;

	if (dockedge >= nextedge) {
		printf("\nERROR: Trying to add leaf %d", nextleaf);
		printf(" to nonexisting edge %d [addnextleaf]!\n", dockedge);
		exit(1);
	}
	
	if (nextleaf >= Maxspc) {
		printf("\nERROR: Trying to add leaf %d", nextleaf);
		printf(" to a tree with %d leaves [addnextleaf]!\n", Maxspc);
		exit(1);
	}
		
	/* necessary change in edgeofleaf if dockedge == edgeofleaf[0] */
	if (edgeofleaf[0] == dockedge) edgeofleaf[0] = nextedge;
	
	/* adding nextedge to the tree */
	edge[nextedge].up = edge[dockedge].up;
	edge[nextedge].downleft = &edge[dockedge];
	edge[nextedge].downright = &edge[nextedge+1];
	edge[dockedge].up = &edge[nextedge];

	if (edge[nextedge].up != NULL) {
		if ( ((edge[nextedge].up)->downleft) == &edge[dockedge] ) 
			(edge[nextedge].up)->downleft  = &edge[nextedge];
		else   
			(edge[nextedge].up)->downright = &edge[nextedge];
	}

	/* adding nextedge + 1 to the tree */
	edge[nextedge+1].up = &edge[nextedge];
	edge[nextedge+1].downleft = NULL;
	edge[nextedge+1].downright = NULL;
	edgeofleaf[nextleaf] = nextedge+1;
	
	/* the two new edges get info about the old edges */
	/* nextedge */
	for (i = 0; i < nextedge; i++) {
		switch ( (edge[dockedge].edgemap)[i] ) {
			
			/* down right changes to down left */
			case 5:		(edge[nextedge].edgemap)[i] = 4;
						break;
						
			/* null changes to down left */
			case 0:		(edge[nextedge].edgemap)[i] = 4;
						break;
			
			default: 	(edge[nextedge].edgemap)[i] =
							(edge[dockedge].edgemap)[i];
						break;
		}
	}

	/* nextedge + 1 */
	for (i = 0; i < nextedge; i++) {
		switch ( (edge[dockedge].edgemap)[i] ) {
			
			/* up/down left changes to up */
			case 2:		(edge[nextedge+1].edgemap)[i] = 1;
						break;
			
			/* up/down right changes to up */		
			case 3:		(edge[nextedge+1].edgemap)[i] = 1;
						break;
						
			/* down left changes to up/down left */			
			case 4:		(edge[nextedge+1].edgemap)[i] = 2;
						break;

			/* down right changes to up/down left */	
			case 5:		(edge[nextedge+1].edgemap)[i] = 2;
						break;
			
			/* null changes to up/down left */
			case 0:		(edge[nextedge+1].edgemap)[i] = 2;
						break;
			
			/* up stays up */
			default: 	(edge[nextedge+1].edgemap)[i] =
							(edge[dockedge].edgemap)[i];
						break;
		}
	}

	/* dockedge */
	for (i = 0; i < nextedge; i++) {
		switch ( (edge[dockedge].edgemap)[i] ) {
			
			/* up/down right changes to up */
			case 3:		(edge[dockedge].edgemap)[i] = 1;
						break;
						
			/* up/down left changes to up */
			case 2:		(edge[dockedge].edgemap)[i] = 1;
						break;
			
			default: 	break;
		}
	}

	/* all edgemaps are updated for the two new edges */
	/* nextedge */
	(edge[nextedge].edgemap)[nextedge] = 0;
	(edge[nextedge].edgemap)[nextedge+1] = 5; /* down right */
	
	/* nextedge + 1 */
	(edge[nextedge+1].edgemap)[nextedge] = 1; /* up */
	(edge[nextedge+1].edgemap)[nextedge+1] = 0;
	
	/* all other edges */
	for (i = 0; i < nextedge; i++) {
		(edge[i].edgemap)[nextedge] = (edge[i].edgemap)[dockedge];
		(edge[i].edgemap)[nextedge+1] = (edge[i].edgemap)[dockedge];
	}
	
	/* an extra for dockedge */
	(edge[dockedge].edgemap)[nextedge] = 1; /* up */
	(edge[dockedge].edgemap)[nextedge+1] = 3; /* up/down right */

	nextleaf++;
	nextedge = nextedge + 2;
}

/* free memory (to be called after inittree) */
void freetree()
{
	int i;

	for (i = 0; i < 2 * Maxspc - 3; i++) free(edge[i].edgemap);
	free(edge);
	free(edgeofleaf);	
}

#ifndef CONTBAYES

/* clear all edgeinfos */
void resetedgeinfo()
{
	int i;
	
	for (i = 0; i < nextedge; i++)
		edge[i].edgeinfo = 0;
}

#else

/* clear all edgescores */
void resetedgescore()
{
	int i;
	
	for (i = 0; i < nextedge; i++)
		edge[i].edgescore = 0.0;
}

#endif

#ifndef CONTBAYES

/* increment all edgeinfo between leaf A and B */
void incrementedgeinfo(int A, int B)
{	
	int curredge, finaledge, nextstep;
	
	if (A == B) return;
	
	finaledge = edgeofleaf[B];
	
	curredge = edgeofleaf[A];
	edge[curredge].edgeinfo = edge[curredge].edgeinfo + 1;
	
	while (curredge != finaledge) {
		nextstep = (edge[curredge].edgemap)[finaledge];
		switch (nextstep) {

			/* up */
			case 1:	curredge = (edge[curredge].up)->numedge;
					break;
				
			/* up/down left */
			case 2:	curredge = ((edge[curredge].up)->downleft)->numedge;
					break;

			/* up/down right */
			case 3:	curredge = ((edge[curredge].up)->downright)->numedge;
					break;
				
			/* down left */
			case 4:	curredge = (edge[curredge].downleft)->numedge;
					break;
				
			/* down right */
			case 5:	curredge = (edge[curredge].downright)->numedge;
					break;
				
		}
		edge[curredge].edgeinfo = edge[curredge].edgeinfo + 1;
	}	
}

#else

/* increment all edgescore between leaf A and B by the amount w */
void incrementedgescore(int A, int B, double w)
{	
	int curredge, finaledge, nextstep;
	
	if (A == B) return;
	
	finaledge = edgeofleaf[B];
	
	curredge = edgeofleaf[A];
	edge[curredge].edgescore = edge[curredge].edgescore + w;
	
	while (curredge != finaledge) {
		nextstep = (edge[curredge].edgemap)[finaledge];
		switch (nextstep) {

			/* up */
			case 1:	curredge = (edge[curredge].up)->numedge;
					break;
				
			/* up/down left */
			case 2:	curredge = ((edge[curredge].up)->downleft)->numedge;
					break;

			/* up/down right */
			case 3:	curredge = ((edge[curredge].up)->downright)->numedge;
					break;
				
			/* down left */
			case 4:	curredge = (edge[curredge].downleft)->numedge;
					break;
				
			/* down right */
			case 5:	curredge = (edge[curredge].downright)->numedge;
					break;
				
		}
		edge[curredge].edgescore = edge[curredge].edgescore + w;
	}	
}

#endif

#ifndef CONTBAYES

/* checks which edge has the lowest edgeinfo
   if there are several edges with the same lowest edgeinfo,
   one of them will be selected randomly */
void minimumedgeinfo()
{
	int i, k, howmany, randomnum;

	howmany = 1;
	minedge = 0;
	mininfo = edge[0].edgeinfo;
	for (i = 1; i < nextedge; i++)
		if (edge[i].edgeinfo <= mininfo) {
			if (edge[i].edgeinfo == mininfo) {
				howmany++;
			} else {
				minedge = i;
				mininfo = edge[i].edgeinfo;
				howmany = 1;
			}
		}
	
	if (howmany > 1) { /* draw random edge */
		randomnum = randominteger(howmany) + 1; /* 1 to howmany */
		i = -1;
		for (k = 0; k < randomnum; k++) {
			do {
				i++;
			} while (edge[i].edgeinfo != mininfo);
			minedge = i;
		}
	}
}

#else

/* check which edge has the lowest edgescore
   if there are several edges with the same lowest edgescore,
   one of them will be selected randomly */
void minimumedgescore()
{
	int i, k, howmany, randomnum;

	howmany = 1;
	minedge = 0;
	minscore = edge[0].edgescore;
	for (i = 1; i < nextedge; i++)
		if (edge[i].edgescore <= minscore) {
			if (edge[i].edgescore == minscore) {
				howmany++;
			} else {
				minedge = i;
				minscore = edge[i].edgescore;
				howmany = 1;
			}
		}
	
	if (howmany > 1) { /* draw random edge */
		randomnum = randominteger(howmany) + 1; /* 1 to howmany */
		i = -1;
		for (k = 0; k < randomnum; k++) {
			do {
				i++;
			} while (edge[i].edgescore != minscore);
			minedge = i;
		}
	}
}

#endif

/******************************************************************************/
/* functions for computing the consensus tree                                 */
/******************************************************************************/

/* prepare for consensus tree analysis */
void initconsensus()
{
	biparts = new_cmatrix(Maxspc-3, Maxspc);
	if (Maxspc % 32 == 0)
		splitlength = Maxspc/32;
	else splitlength = (Maxspc + 32 - (Maxspc % 32))/32;
	numbiparts = 0; /* no pattern stored so far */
	maxbiparts = 0; /* no memory reserved so far */
	splitfreqs = NULL;
	splitpatterns = NULL;
	splitsizes = NULL;
	splitcomp = (ulint *) malloc(splitlength * sizeof(ulint) );
	if (splitcomp == NULL) maerror("splitcomp");
}

/* prototype needed for recursive function */
void makepart(int i, int curribrnch);

/* recursive function to get bipartitions */
void makepart(int i, int curribrnch)
{	
	int j;
	
	if ( edge[i].downright == NULL ||
		  edge[i].downleft == NULL) { /* if i is leaf */
			
			/* check out what leaf j sits on this edge i */		
			for (j = 1; j < Maxspc; j++) {
				if (edgeofleaf[j] == i) {
					biparts[curribrnch][trueID[j]] = '*';	
					return;
				}	
			}
	} else { /* still on inner branch */
		makepart(edge[i].downleft->numedge, curribrnch);
		makepart(edge[i].downright->numedge, curribrnch);
	}
}

/* compute bipartitions of tree of current puzzling step */
void computebiparts()
{
	int i, j, curribrnch;
	
	curribrnch = -1;
	
	for (i = 0; i < Maxspc - 3; i++)
		for (j = 0; j < Maxspc; j++)
			biparts[i][j] = '.';

	for (i = 0; i < Maxbrnch; i++) {
		if (!(     edgeofleaf[0] == i    ||
		       edge[i].downright == NULL ||
		        edge[i].downleft == NULL) ) { /* check all inner branches */
			curribrnch++;
			makepart(i, curribrnch);
			
			/* make sure that the root is always a '*' */
			if (biparts[curribrnch][outgroup] == '.') {
				for (j = 0; j < Maxspc; j++) {
					if (biparts[curribrnch][j] == '.')
						biparts[curribrnch][j] = '*';
					else
						biparts[curribrnch][j] = '.';
				}
			}
		}
	}
}

/* print out the bipartition n of all different splitpatterns */
void printsplit(FILE *fp, ulint n)
{
	int i, j, col;
	ulint z;
	
	col = 0;
	for (i = 0; i < splitlength; i++) {
		z = splitpatterns[n*splitlength + i];
		for (j = 0; j < 32 && col < Maxspc; j++) {
			if (col % 10 == 0 && col != 0) fprintf(fp, " ");
			if (z & 1) fprintf(fp, ".");
			else fprintf(fp, "*");
			z = (z >> 1);
			col++;
		}
	}
}

/* private realloc function that works also on non-ANSI compilers */ 
void *myrealloc(void *p, size_t s)
{
	if (p == NULL) return malloc(s);
	else return realloc(p, s);
}

/* make new entries for new different bipartitions and count frequencies */
void makenewsplitentries()
{
	int i, j, bpc, identical, idflag, bpsize;
	ulint nextentry, obpc;

	/* where the next entry would be in splitpatterns */
	nextentry = numbiparts;
	
	for (bpc = 0; bpc < Maxspc - 3; bpc++) { /* for every new bipartition */	
		/* convert bipartition into a more compact format */
		bpsize = 0;
		for (i = 0; i < splitlength; i++) {
			splitcomp[i] = 0;	
			for (j = 0; j < 32; j++) {
				splitcomp[i] = splitcomp[i] >> 1;
				if (i*32 + j < Maxspc)
					if (biparts[bpc][i*32 + j] == '.') {
						/* set highest bit */
						splitcomp[i] = (splitcomp[i] | 2147483648);
						bpsize++; /* count the '.' */
					}
			}
		}		
		/* compare to the *old* patterns */
		identical = FALSE;
		for (obpc = 0; (obpc < numbiparts) && (!identical); obpc++) {
			/* compare first partition size */
			if (splitsizes[obpc] == bpsize) idflag = TRUE;
			else idflag = FALSE;
			/* if size is identical compare whole partition */
			for (i = 0; (i < splitlength) && idflag; i++)
				if (splitcomp[i] != splitpatterns[obpc*splitlength + i])
					idflag = FALSE;
			if (idflag) identical = TRUE;
		}
		if (identical) { /* if identical increase frequency */
			splitfreqs[2*(obpc-1)]++;
		} else { /* create new entry */
			if (nextentry == maxbiparts) { /* reserve more memory */
				maxbiparts = maxbiparts + 2*Maxspc;
				splitfreqs = (ulint *) myrealloc(splitfreqs,
					2*maxbiparts * sizeof(ulint) );
				/* 2x: splitfreqs contains also an index (sorting!) */
				if (splitfreqs == NULL) maerror("splitfreqs");
				splitpatterns = (ulint *) myrealloc(splitpatterns,
					splitlength*maxbiparts * sizeof(ulint) );
				if (splitpatterns == NULL) maerror("splitpatterns");
				splitsizes = (int *) myrealloc(splitsizes,
					maxbiparts * sizeof(int) );
				if (splitsizes == NULL) maerror("splitsizes");
			}
			splitfreqs[2*nextentry] = 1; /* frequency */
			splitfreqs[2*nextentry+1] = nextentry; /* index for sorting */
			for (i = 0; i < splitlength; i++)
				splitpatterns[nextentry*splitlength + i] = splitcomp[i];
			splitsizes[nextentry] = bpsize;
			nextentry++;
		}
	}
	numbiparts = nextentry;
}

/* general remarks:

   - every entry in consbiparts is one node of the consensus tree
   - for each node one has to know which taxa and which other nodes
     are *directly* descending from it
   - for every taxon/node number there is a flag that shows
     whether it descends from the node or not
   - '0' means that neither a taxon nor another node with the
         corresponding number decends from the node
     '1' means that the corresponding taxon descends from the node
     '2' means that the corresponding node descends from the node
     '3' means that the corresponding taxon and node descends from the node
*/

/* copy bipartition n of all different splitpatterns to consbiparts[k] */
void copysplit(ulint n, int k)
{
	int i, j, col;
	ulint z;
	
	col = 0;
	for (i = 0; i < splitlength; i++) {
		z = splitpatterns[n*splitlength + i];
		for (j = 0; j < 32 && col < Maxspc; j++) {
			if (z & 1) consbiparts[k][col] = '1';
			else consbiparts[k][col] = '0';
			z = (z >> 1);
			col++;
		}
	}
}

/* compute majority rule consensus tree */
void makeconsensus()
{
	int i, j, k, size, subnode;
	char chari, charj;

	/* sort bipartition frequencies */
	qsort(splitfreqs, numbiparts, 2*sizeof(ulint), ulintcmp);
	/* how many bipartitions are included in the consensus tree */
	consincluded = 0;
	for (i = 0; i < numbiparts && i == consincluded; i++) {
		if (2*splitfreqs[2*i] > Numtrial) consincluded = i + 1;
	}

#ifdef SIMULATION
consincluded = 5;
#endif

	/* collect all info about majority rule consensus tree */
	/* the +1 is due to the edge with the root */
	consconfid = new_ivector(consincluded + 1);
	conssizes = new_ivector(2*consincluded + 2);
	consbiparts = new_cmatrix(consincluded + 1, Maxspc);
	
	for (i = 0; i < consincluded; i++) {
		/* copy partition to consbiparts */
		copysplit(splitfreqs[2*i+1], i);
		/* frequency in percent (rounded to integer) */
		consconfid[i] = (int) floor(100.0*splitfreqs[2*i]/Numtrial + 0.5);
		/* size of partition */
		conssizes[2*i] = splitsizes[splitfreqs[2*i+1]];
		conssizes[2*i+1] = i;
	}
	for (i = 0; i < Maxspc; i++) consbiparts[consincluded][i] = '1';
	consbiparts[consincluded][outgroup] = '0';
	consconfid[consincluded] = 100;
	conssizes[2*consincluded] = Maxspc - 1;
	conssizes[2*consincluded + 1] = consincluded;

	/* sort bipartitions according to cluster size */
	qsort(conssizes, consincluded + 1, 2*sizeof(int), intcmp);

	/* reconstruct consensus tree */
	for (i = 0; i < consincluded; i++) { /* try every node */
		size = conssizes[2*i]; /* size of current node */
		for (j = i + 1; j < consincluded + 1; j++) {
		
			/* compare only with nodes with more descendants */
			if (size == conssizes[2*j]) continue;
			
			/* check whether node i is a subnode of j */
			subnode = FALSE;
			for (k = 0; k < Maxspc && !subnode; k++) {
				chari = consbiparts[ conssizes[2*i+1] ][k];
				if (chari != '0') {
					charj = consbiparts[ conssizes[2*j+1] ][k];
					if (chari == charj || charj == '3') subnode = TRUE;
				}
			}
			
			/* if i is a subnode of j change j accordingly */
			if (subnode) {
				/* remove subnode i from j */
				for (k = 0; k < Maxspc; k++) {
					chari = consbiparts[ conssizes[2*i+1] ][k];
					if (chari != '0') {
						charj = consbiparts[ conssizes[2*j+1] ][k];
						if (chari == charj)
							consbiparts[ conssizes[2*j+1] ][k] = '0';
						else if (charj == '3') {
							if (chari == '1')
								consbiparts[ conssizes[2*j+1] ][k] = '2';
							else if (chari == '2')
								consbiparts[ conssizes[2*j+1] ][k] = '1';
							else {
								printf("\nERROR: Consensus tree [1]\n");
								exit(1);
							}	
						} else {
							printf("\nERROR: Consensus tree [2]\n");
							exit(1);
						}
					}
				}
				/* add link to subnode i in node j */
				charj = consbiparts[ conssizes[2*j+1] ][ conssizes[2*i+1] ];
				if (charj == '0')
					consbiparts[ conssizes[2*j+1] ][ conssizes[2*i+1] ] = '2';
				else if (charj == '1')
					consbiparts[ conssizes[2*j+1] ][ conssizes[2*i+1] ] = '3';
				else {
					printf("\nERROR: Consensus tree [3]\n");
					exit(1);
				}
			}
		}
	}
}

/* prototype for recursion */
void writenode(FILE *treefile, int node);

/* write node (writeconsensustree) */
void writenode(FILE *treefile, int node)
{
	int i, first;
	
	fprintf(treefile, "(");
	column++;
	/* write descending nodes */
	first = TRUE;
	for (i = 0; i < Maxspc; i++) {
		if (consbiparts[node][i] == '2' ||
		consbiparts[node][i] == '3') {
			if (first) first = FALSE;
			else {
				fprintf(treefile, ",");
				column++;
			}
			if (column > 60) {
				column = 2;
				fprintf(treefile, "\n");
			}
			/* write node i */
			writenode(treefile, i);

#ifdef SIMULATION
fprintf(treefile, ":%d", consconfid[i]);
#else
			/* reliability value as internal label */
			fprintf(treefile, "%d", consconfid[i]);
#endif

			column = column + 3;
		}
	}
	/* write descending taxa */
	for (i = 0; i < Maxspc; i++) {
		if (consbiparts[node][i] == '1' ||
		consbiparts[node][i] == '3') {
			if (first) first = FALSE;
			else {
				fprintf(treefile, ",");
				column++;
			}
			if (column > 60) {
				column = 2;
				fprintf(treefile, "\n");
			}
			column += fputid(treefile, i);
		}
	}
	fprintf(treefile, ")");
	column++;
}

/* write consensus tree */
void writeconsensustree(FILE *treefile)
{
	int i, first;
	
	column = 1;
	fprintf(treefile, "(");
	column += fputid(treefile, outgroup) + 2;
	fprintf(treefile, ",");
	/* write descending nodes */
	first = TRUE;
	for (i = 0; i < Maxspc; i++) {
		if (consbiparts[consincluded][i] == '2' ||
		consbiparts[consincluded][i] == '3') {
			if (first) first = FALSE;
			else {
				fprintf(treefile, ",");
				column++;
			}
			if (column > 60) {
				column = 2;
				fprintf(treefile, "\n");
			}
			/* write node i */
			writenode(treefile, i);

#ifdef SIMULATION
fprintf(treefile, ":%d", consconfid[i]);
#else
			/* reliability value as internal label */
			fprintf(treefile, "%d", consconfid[i]);
#endif

			column = column + 3;
		}
	}
	/* write descending taxa */
	for (i = 0; i < Maxspc; i++) {
		if (consbiparts[consincluded][i] == '1' ||
		consbiparts[consincluded][i] == '3') {
			if (first) first = FALSE;
			else {
				fprintf(treefile, ",");
				column++;
			}
			if (column > 60) {
				column = 2;
				fprintf(treefile, "\n");
			}
			column += fputid(treefile, i);
		}
	}
	fprintf(treefile, ");\n");
}

/* prototype for recursion */
void nodecoordinates(int node);

/* establish node coordinates (plotconsensustree) */
void nodecoordinates(int node)
{
	int i, ymin, ymax, xcoordinate;

	/* first establish coordinates of descending nodes */
	for (i = 0; i < Maxspc; i++) {
		if (consbiparts[node][i] == '2' ||
		consbiparts[node][i] == '3') 
			nodecoordinates(i);
	}
	
	/* then establish coordinates of descending taxa */
	for (i = 0; i < Maxspc; i++) {
		if (consbiparts[node][i] == '1' ||
		consbiparts[node][i] == '3') {
			/* y-coordinate of taxon i */
			ycortax[i] = ytaxcounter;
			ytaxcounter = ytaxcounter - 2;
		}
	}
	
	/* then establish coordinates of this node */
	ymin = 2*Maxspc - 2;
	ymax = 0;
	xcoordinate = 0;
	for (i = 0; i < Maxspc; i++) {
		if (consbiparts[node][i] == '2' ||
		consbiparts[node][i] == '3') {
			if (ycor[i] > ymax) ymax = ycor[i];
			if (ycor[i] < ymin) ymin = ycor[i];
			if (xcor[i] > xcoordinate) xcoordinate = xcor[i];
		}
	}
	for (i = 0; i < Maxspc; i++) {
		if (consbiparts[node][i] == '1' ||
		consbiparts[node][i] == '3') {
			if (ycortax[i] > ymax) ymax = ycortax[i];
			if (ycortax[i] < ymin) ymin = ycortax[i];
		}
	}
	ycormax[node] = ymax;
	ycormin[node] = ymin;
	ycor[node] = (int) floor(0.5*(ymax + ymin) + 0.5);
	if (xcoordinate == 0) xcoordinate = 9;
	xcor[node] = xcoordinate + 4;
}

/* prototype for recursion */
void drawnode(int node, int xold);

/* drawnode  (plotconsensustree) */
void drawnode(int node, int xold)
{
	int i, j;
	char buf[4];
	
	/* first draw vertical line */
	for (i = ycormin[node] + 1; i < ycormax[node]; i++)
		treepict[xcor[node]][i] = ':';
		
	/* then draw descending nodes */
	for (i = 0; i < Maxspc; i++) {
		if (consbiparts[node][i] == '2' ||
		consbiparts[node][i] == '3') 
			drawnode(i, xcor[node]);
	}
	
	/* then draw descending taxa */
	for (i = 0; i < Maxspc; i++) {
		if (consbiparts[node][i] == '1' ||
		consbiparts[node][i] == '3') {
			treepict[xcor[node]][ycortax[i]] = ':';
			for (j = xcor[node] + 1; j < xsize-10; j++)
				treepict[j][ycortax[i]] = '-';
			for (j = 0; j < 10; j++)
				treepict[xsize-10+j][ycortax[i]] = Identif[i][j];	
		}
	}
	
	/* then draw internal edge with consensus value */
	treepict[xold][ycor[node]] = ':';
	treepict[xcor[node]][ycor[node]] = ':';
	for (i = xold + 1; i < xcor[node]-3; i++)
		treepict[i][ycor[node]] = '-';
	sprintf(buf, "%d", consconfid[node]);
	if (consconfid[node] == 100) {
		treepict[xcor[node]-3][ycor[node]] = buf[0];
		treepict[xcor[node]-2][ycor[node]] = buf[1];
		treepict[xcor[node]-1][ycor[node]] = buf[2];	
	} else {
		treepict[xcor[node]-3][ycor[node]] = '-';
		treepict[xcor[node]-2][ycor[node]] = buf[0];
		treepict[xcor[node]-1][ycor[node]] = buf[1];
	}
}

/* plot consensus tree */
void plotconsensustree(FILE *plotfp)
{
	int i, j, yroot;
	
	/* x-y-coordinates of each bipartition */
	xcor = new_ivector(consincluded);
	ycor = new_ivector(consincluded);
	ycormax = new_ivector(consincluded);
	ycormin = new_ivector(consincluded);
	/* y-coordinates of each taxon */
	ycortax = new_ivector(Maxspc);
	ycortax[outgroup] = 0;
	
	/* establish coordinates */
	ytaxcounter = 2*Maxspc - 2;
	
	/* first establish coordinates of descending nodes */
	for (i = 0; i < Maxspc; i++) {
		if (consbiparts[consincluded][i] == '2' ||
		consbiparts[consincluded][i] == '3') 
			nodecoordinates(i);
	}
	
	/* then establish coordinates of descending taxa */
	for (i = 0; i < Maxspc; i++) {
		if (consbiparts[consincluded][i] == '1' ||
		consbiparts[consincluded][i] == '3') {
			/* y-coordinate of taxon i */
			ycortax[i] = ytaxcounter;
			ytaxcounter = ytaxcounter - 2;
		}
	}

	/* then establish length of root edge and size of whole tree */
	yroot = 0;
	xsize = 0;
	for (i = 0; i < Maxspc; i++) {
		if (consbiparts[consincluded][i] == '2' ||
		consbiparts[consincluded][i] == '3') {
			if (ycor[i] > yroot) yroot = ycor[i];
			if (xcor[i] > xsize) xsize = xcor[i];
		}
	}
	for (i = 0; i < Maxspc; i++) {
		if (consbiparts[consincluded][i] == '1' ||
		consbiparts[consincluded][i] == '3') {
			if (ycortax[i] > yroot) yroot = ycortax[i];
		}
	}
	if (xsize == 0) xsize = 9;
	/* size in x direction inclusive one blank on the left */
	xsize = xsize + 6; 
	
	/* change all x-labels so that (0,0) is down-left */
	for (i = 0; i < consincluded; i++)
		xcor[i] = xsize-1-xcor[i];
	
	/* draw tree */
	treepict = new_cmatrix(xsize, 2*Maxspc-1);
	for (i = 0; i < xsize; i++)
		for (j = 0; j < 2*Maxspc-1; j++)
			treepict[i][j] = ' ';
	
	/* draw root */
	for (i = 1; i < yroot; i++)
		treepict[1][i] = ':';
	treepict[1][0] = ':';
	for (i = 2; i < xsize - 10; i++)
		treepict[i][0] = '-';
	for (i = 0; i < 10; i++)
		treepict[xsize-10+i][0] = Identif[outgroup][i];
	
	/* then draw descending nodes */
	for (i = 0; i < Maxspc; i++) {
		if (consbiparts[consincluded][i] == '2' ||
		consbiparts[consincluded][i] == '3') 
			drawnode(i, 1);
	}
	
	/* then draw descending taxa */
	for (i = 0; i < Maxspc; i++) {
		if (consbiparts[consincluded][i] == '1' ||
		consbiparts[consincluded][i] == '3') {
			treepict[1][ycortax[i]] = ':';
			for (j = 2; j < xsize-10; j++)
				treepict[j][ycortax[i]] = '-';
			for (j = 0; j < 10; j++)
				treepict[xsize-10+j][ycortax[i]] = Identif[i][j];	
		}
	}
	
	/* plot tree */
	for (i = 2*Maxspc-2; i > -1; i--) {
		for (j = 0; j < xsize; j++)
			fputc(treepict[j][i], plotfp);
		fputc('\n', plotfp);
	}	
	
	free_ivector(xcor);
	free_ivector(ycor);
	free_ivector(ycormax);
	free_ivector(ycormin);
	free_ivector(ycortax);
	free_cmatrix(treepict);
}


/******************************************************************************/
/* storing and evaluating quartet branching information                       */
/******************************************************************************/

/* general remarks:

	for a quartet with the taxa a, b, c, d there are
	three possible binary trees:
	
		1)  (a,b)-(c,d)
		2)  (a,c)-(b,d)
		3)  (a,d)-(b,c)
	
	For every quartet information about its branching structure is
	stored. With the functions  readquartet  and  writequartet
	this information can be accessed. For every quartet (a,b,c,d)
	with a < b < c < d (taxa) the branching information is encoded
	using 4 bits:
	
	value          8             4             2             1
	        +-------------+-------------+-------------+-------------+
	        |  not used   |   tree 3    |   tree 2    |   tree 1    |
	        +-------------+-------------+-------------+-------------+

	If the branching structure of the taxa corresponds to one of the
	three trees the corresponding bit is set. If the branching structure
	is unclear because two of the three trees have the same maximum
	likelihood value the corresponding two bits are set. If the branching
	structure is completely unknown all the bits are set (the highest
	bit is always cleared because it is not used).

*/

#ifndef CONTBAYES

/* allocate memory for quartets */
void mallocquartets(int taxa)
{
	ulint nc, numch;

	/* number of quartets */
	Numquartets = (ulint) taxa*(taxa-1)*(taxa-2)*(taxa-3)/24;
	if (Numquartets % 2 == 0) { /* even number */
		numch = Numquartets/2;
	} else { /* odd number */
		numch = (Numquartets + 1)/2;
	}
	/* allocate memory */
	quartetinfo = (unsigned char *) malloc(numch * sizeof(unsigned char) );
	if (quartetinfo == NULL) maerror("mallocquartets");
	for (nc = 0; nc < numch; nc++) quartetinfo[nc] = 0;
}

/* free quartet memory */
void freequartets()
{	
	free(quartetinfo);
}

#else

/* allocate memory for continous Bayesian weights */
void mallocquartetweights(int taxa)
{
	/* number of quartets */
	Numquartets = (ulint) taxa*(taxa-1)*(taxa-2)*(taxa-3)/24;
	
	/* allocate memory */
	quartetw1 = (double *) malloc(Numquartets * sizeof(double) );
	if (quartetw1 == NULL) maerror("mallocquartetsweights (w1)");
	quartetw2 = (double *) malloc(Numquartets * sizeof(double) );
	if (quartetw2 == NULL) maerror("mallocquartetsweights (w2)");
	quartetw3 = (double *) malloc(Numquartets * sizeof(double) );
	if (quartetw3 == NULL) maerror("mallocquartetsweights (w3)");
}

/* free memory of Bayesian weights */
void freequartetweights()
{	
	free(quartetw1);
	free(quartetw2);
	free(quartetw3);
}

#endif

#ifndef CONTBAYES

/* read quartet info - a < b < c < d */
unsigned char readquartet(int a, int b, int c, int d)
{
	ulint qnum;

	qnum = (ulint) a
			+ (ulint) b*(b-1)/2
			+ (ulint) c*(c-1)*(c-2)/6
			+ (ulint) d*(d-1)*(d-2)*(d-3)/24;
	if (qnum % 2 == 0) { /* even number */
		/* bits 0 to 3 */
		return (quartetinfo[qnum/2] & (unsigned char) 15);
	} else { /* odd number */
		/* bits 4 to 7 */
		return ((quartetinfo[(qnum-1)/2] & (unsigned char) 240)>>4);
	}
}

/* write quartet info - a < b < c < d, 0 <= info <= 15 */
void writequartet(int a, int b, int c, int d, unsigned char info)
{
	ulint qnum;

	qnum = (ulint) a
			+ (ulint) b*(b-1)/2
			+ (ulint) c*(c-1)*(c-2)/6
			+ (ulint) d*(d-1)*(d-2)*(d-3)/24;
	if (qnum % 2 == 0) { /* even number */
		/* bits 0 to 3 */
		quartetinfo[qnum/2] =
			((quartetinfo[qnum/2] & (unsigned char) 240) |
			(info & (unsigned char) 15));
	} else { /* odd number */
		/* bits 4 to 7 */
		quartetinfo[(qnum-1)/2] =
			((quartetinfo[(qnum-1)/2] & (unsigned char) 15) |
			((info & (unsigned char) 15)<<4));
	}
}

#else

/* read quartet weights - a < b < c < d */
void readquartetweights(int a, int b, int c, int d, double *w1, double *w2, double *w3)
{
	ulint qnum;

	qnum = (ulint) a
			+ (ulint) b*(b-1)/2
			+ (ulint) c*(c-1)*(c-2)/6
			+ (ulint) d*(d-1)*(d-2)*(d-3)/24;
	*w1 = quartetw1[qnum];
	*w2 = quartetw2[qnum];
	*w3 = quartetw3[qnum];
}

/* write quartet weights - a < b < c < d */
void writequartetweights(int a, int b, int c, int d, double w1, double w2, double w3)
{
	ulint qnum;

	qnum = (ulint) a
			+ (ulint) b*(b-1)/2
			+ (ulint) c*(c-1)*(c-2)/6
			+ (ulint) d*(d-1)*(d-2)*(d-3)/24;
	quartetw1[qnum] = w1;
	quartetw2[qnum] = w2;
	quartetw3[qnum] = w3;
}

#endif

/* prototypes */
void openfiletowrite(FILE **fp, char name[]);
void closefile(FILE *fp);

/* checks out all possible quartets */
void computeallquartets()
{	
	int a, b, c, i, badquartet;
	double d1, d2, d3, qc2, mintogo, minutes, hours, temp, wlist[6], plist[6];
	unsigned char qpbranching;
	ulint nq;

	if (tri_optn) { /* generate characteristic triangle */
		openfiletowrite(&trifp, TRIANGLE);
		fprintf(trifp, "(* characteristic triangle *)\n\n");
		fprintf(trifp, "p1 = {0,0};\n");
		fprintf(trifp, "p2 = {1,0};\n");
		fprintf(trifp, "p3 = {1/2, 1/2 Sqrt[3]};\n");
		fprintf(trifp, "triangle = {Line[{p1, p2}], Line[{p2, p3}], Line[{p3, p1}]};\n");
		fprintf(trifp, "\nShow[Graphics[\n");
		fprintf(trifp, "	{PointSize[0.003],\n");
	}
	
	if (show_optn) { /* list all unresolved quartets */
		openfiletowrite(&unresfp, UNRESOLVED);
		fprintf(unresfp, "List of all completely unresolved quartets:\n\n");
	}

	nq = 0;
	badqs = 0;
	
	/* start timer - percentage of completed quartets */
	time(&time0);
	time1 = time0;
	
	for (i = 3; i < Maxspc; i++) 
			for (a = 0; a < i - 2; a++)
					for (b = a + 1; b < i - 1; b++)
						for (c = b + 1; c < i; c++) {
							nq++;

							/* generate message every 15 minutes */
							/* check timer */
							time(&time2);
							if ( (time2 - time1) > 900) {
								/* every 900 seconds */
								/* percentage of completed quartets */
								qc2 = 100.*nq/Numquartets;
								mintogo = (100.0-qc2) *
									(double) (time2-time0)/60.0/qc2;
								hours = floor(mintogo/60.0);
								minutes = mintogo - 60.0*hours;
								printf("\n        %.2f%%", qc2);
								printf(" of all quartets");
								printf(" completed (remaining");
								printf(" time: %.0f", hours);
								printf(" hours %.0f", minutes);
								printf(" minutes)\n");
								time1 = time2;
							}
							
							/* maximum likelihood values */
							/* approximation */
							d1 = quartet_alklhd(a,b,c,i); /* (a,b)-(c,i) */
							d2 = quartet_alklhd(a,c,b,i); /* (a,c)-(b,i) */
							d3 = quartet_alklhd(a,i,b,c); /* (a,i)-(b,c) */

							badquartet = FALSE;

							/* compute Bayesian weights */
							temp = (d1 + d2 + d3)/3.0;
							d1 = exp(d1 - temp);
							d2 = exp(d2 - temp);
							d3 = exp(d3 - temp);
							temp = d1 + d2 + d3;
							wlist[0] = d1 / temp;
							wlist[1] = 1.0;
							wlist[2] = d2 / temp;
							wlist[3] = 2.0;
							wlist[4] = d3 / temp;
							wlist[5] = 4.0;

#ifdef CONTBAYES
							writequartetweights(a, b, c, i,
								wlist[0], wlist[2], wlist[4]);
#endif
						
							/* generate one point in characteristic triangle */
							if (tri_optn)
								fprintf(trifp,"	Point[ {%.9f, %.9f} ],\n",
								0.5*wlist[0] + wlist[2], wlist[0]*0.8660254038);
							
							/* sort in descending order */
							qsort(wlist, 3, 2*sizeof(double), dcmp);							
							
							/* check out the three possibilities */

							/* 100 distribution */
							plist[0] = (1.0 - wlist[0])*(1.0 - wlist[0]) +
							           (0.0 - wlist[2])*(0.0 - wlist[2]) +
							           (0.0 - wlist[4])*(0.0 - wlist[4]);
							plist[1] = wlist[1];

							/* 110 distribution */
							plist[2] = (0.5 - wlist[0])*(0.5 - wlist[0]) +
							           (0.5 - wlist[2])*(0.5 - wlist[2]) +
							           (0.0 - wlist[4])*(0.0 - wlist[4]);
							plist[3] = wlist[1] + wlist[3];

							/* 111 distribution */
							temp = 1.0/3.0;
							plist[4] = (temp - wlist[0])*(temp - wlist[0]) +
							           (temp - wlist[2])*(temp - wlist[2]) +
							           (temp - wlist[4])*(temp - wlist[4]);
							plist[5] = wlist[1] + wlist[3] + wlist[5];

							/* sort in descending order */
							qsort(plist, 3, 2*sizeof(double), dcmp);
			
							qpbranching = (unsigned char) plist[5];

#ifndef CONTBAYES							
							writequartet(a, b, c, i, qpbranching);
#endif							
							/* a bad quartet is a quartet that shows
							   equal weights for all three possible topologies */
							if (qpbranching == 7) badquartet = TRUE;

							if (badquartet) {
								badqs++;
								if (show_optn) {
									fputid10(unresfp, a);
									fprintf(unresfp, "  ");
									fputid10(unresfp, b);
									fprintf(unresfp, "  ");
									fputid10(unresfp, c);
									fprintf(unresfp, "  ");
									fputid(unresfp, i);
									fprintf(unresfp, "\n");
								}
							}
						}
	printf(
		"\nINFO:  %lu quartets analysed (%lu quartets unresolved = %.1f%%)\n",
		nq, badqs, (double) badqs / (double) nq * 100);
	
	/* finish characteristic triangle */
	if (tri_optn) {
		fprintf(trifp, "	triangle\n");
		fprintf(trifp, "	}\n");
		fprintf(trifp, "], PlotRange -> {{0, 1}, {0,1}}, AspectRatio -> 1]\n");
		closefile(trifp);
		
		printf(
		"\nINFO:  Characteristic triangle written to file (%s)\n", TRIANGLE);
	}
	
	if (show_optn) {
		closefile(unresfp);
		printf(
		"\nINFO:  List of all unresolved quartets written to file (%s)\n", UNRESOLVED);
	}
}
							
#ifndef CONTBAYES

/* check the branching structure between the leaves (not the taxa!)
   A, B, C, and I (A, B, C, I don't need to be ordered). As a result,
   the two leaves that are closer related to each other than to leaf I
   are found in chooseA and chooseB. If the branching structure is
   not uniquely defined, ChooseA and ChooseB are chosen randomly
   from the possible taxa */
void checkquartet(int A, int B, int C, int I)
{
	int i, j, a, b, taxon[5], leaf[5], ipos;
	unsigned char qresult;
	int notunique;

	/* The relationship between leaves and taxa is defined by trueID */
	taxon[1] = trueID[A]; /* taxon number */
	leaf[1] = A;          /* leaf number  */
	taxon[2] = trueID[B];
	leaf[2] = B;
	taxon[3] = trueID[C];
	leaf[3] = C;
	taxon[4] = trueID[I];
	leaf[4] = I;

	/* sort after taxa */
	/* Source: Numerical Recipes (PIKSR2.C) */
	for (j = 2; j <= 4; j++) {
		a = taxon[j];
		b = leaf[j];
		i = j-1;
		while (i > 0 && taxon[i] > a) {
			taxon[i+1] = taxon[i];
			leaf[i+1] = leaf[i];
			i--;
		}
		taxon[i+1] = a;
		leaf[i+1] = b;
	}

	/* where is leaf I ? */
	ipos = 1;
	while (leaf[ipos] != I) ipos++;

	/* look at taxa quartet */
	qresult = readquartet(taxon[1], taxon[2], taxon[3], taxon[4]);

	/* chooseA and chooseB */
	do {	
		switch (qresult) {
		
			/* one single branching structure */
		
			/* 001 */
			case 1:		if (ipos == 1 || ipos == 2) {
							chooseA = leaf[3];
							chooseB = leaf[4];
						} else {
							chooseA = leaf[1];
							chooseB = leaf[2];
						}
						notunique = FALSE;
						break;

			/* 010 */
			case 2:		if (ipos == 1 || ipos == 3) {
							chooseA = leaf[2];
							chooseB = leaf[4];
						} else {
							chooseA = leaf[1];
							chooseB = leaf[3];
						}
						notunique = FALSE;
						break;

			/* 100 */
			case 4:		if (ipos == 1 || ipos == 4) {
							chooseA = leaf[2];
							chooseB = leaf[3];
						} else {
							chooseA = leaf[1];
							chooseB = leaf[4];
						}
						notunique = FALSE;
						break;

			/* two possible branching structures */		

			/* 011 */
			case 3:		if (randominteger(2)) qresult = 1;
						else qresult = 2;
						notunique = TRUE;
						break;

			/* 101 */
			case 5:		if (randominteger(2)) qresult = 1;
						else qresult = 4;
						notunique = TRUE;
						break;

			/* 110 */
			case 6:		if (randominteger(2)) qresult = 2;
						else qresult = 4;
						notunique = TRUE;
						break;

			/* three possible branching structures */

			/* 111 */
			case 7:		qresult = (1 << randominteger(3)); /* 1, 2, or 4 */
						notunique = TRUE;
						break;

			default:	printf(
							"\nERROR: Program error [checkquartet]!\n");
						exit(1);
		}
	} while (notunique);

	return;
}

#else

/* Scores equal to Bayesian weights are assigned to the edges on the subtree
   for each of the three weights along the corresponding connection paths */
void assignscores(int A, int B, int C, int I)
{
	int i, j, a, b, taxon[5], leaf[5], ipos;
	unsigned char qresult;
	int notunique;
	double w1, w2, w3;

	/* The relationship between leaves and taxa is defined by trueID */
	taxon[1] = trueID[A]; /* taxon number */
	leaf[1] = A;          /* leaf number  */
	taxon[2] = trueID[B];
	leaf[2] = B;
	taxon[3] = trueID[C];
	leaf[3] = C;
	taxon[4] = trueID[I];
	leaf[4] = I;

	/* sort after taxa */
	/* Source: Numerical Recipes (PIKSR2.C) */
	for (j = 2; j <= 4; j++) {
		a = taxon[j];
		b = leaf[j];
		i = j-1;
		while (i > 0 && taxon[i] > a) {
			taxon[i+1] = taxon[i];
			leaf[i+1] = leaf[i];
			i--;
		}
		taxon[i+1] = a;
		leaf[i+1] = b;
	}

	/* where is leaf I ? */
	ipos = 1;
	while (leaf[ipos] != I) ipos++;

	/* get Bayesian weights */
	readquartetweights(taxon[1], taxon[2], taxon[3], taxon[4], &w1, &w2, &w3);

	/* assign scores for topology (taxon1,taxon2)-(taxon3,taxon4) with weight w1 */
	if (ipos == 1 || ipos == 2)
		incrementedgescore(leaf[3], leaf[4], w1);
	else
		incrementedgescore(leaf[1], leaf[2], w1);

	/* assign scores for topology (taxon1,taxon3)-(taxon2,taxon4) with weight w2 */
	if (ipos == 1 || ipos == 3)
		incrementedgescore(leaf[2], leaf[4], w2);
	else
		incrementedgescore(leaf[1], leaf[3], w2);

	/* assign scores for topology (taxon1,taxon4)-(taxon2,taxon3) with weight w3 */
	if (ipos == 1 || ipos == 4)
		incrementedgescore(leaf[2], leaf[3], w3);
	else
		incrementedgescore(leaf[1], leaf[4], w3);

	return;
}

#endif

/******************************************************************************/
/* sequences                                                                  */
/******************************************************************************/

/* read ten characters of current line as identifier */
void readid(FILE *infp, int t)
{
	int i, j, flag;

	for (i = 0; i < 10; i++) {
		Identif[t][i] = fgetc(infp);
		if (Identif[t][i] == EOF || !isprint(Identif[t][i]) ) {
			printf("\nERROR: Missing full taxon name");
			printf(" (taxon # %d)!\n", t + 1);
			exit(1);
		}
	}	
	/* convert leading blanks in taxon name to underscores */
	flag = FALSE;
	for (i = 9; i > -1; i--) {
		if (flag == FALSE) {
			if (Identif[t][i] != ' ') flag = TRUE; 
		} else {
			if (Identif[t][i] == ' ') Identif[t][i] = '_';
		}
	}
	/* check whether this name is already used */
	for (i = 0; i < t; i++) { /* compare with all other taxa */
		flag = TRUE; /* assume identity */
		for (j = 0; (j < 10) && (flag == TRUE); j++)
			if (Identif[t][j] != Identif[i][j])
				flag = FALSE;
		if (flag) {
			printf("\nERROR: Multiple occurence of taxon name [");
			fputid(stdout, t);
			printf("]\n");
			exit(1);
		}
	}
}

/* read next allowed character */
char readnextcharacter(FILE *ifp, int notu, int nsite)
{
	char c;

	/* ignore blanks and control characters except newline */
	do {
		if (fscanf(ifp, "%c", &c) != 1) {
			printf("\nERROR: Missing character");
			printf(" (taxon # %d [", notu + 1);
			fputid(stdout, notu);
			printf("], ");
			printf("character # %d)!\n", nsite + 1);
			exit(1);
		}
	} while (c == ' ' || (iscntrl(c) && c != '\n'));
	return c;
}

/* skip rest of the line */
void skiprestofline(FILE* ifp, int notu, int nsite)
{
	char c;

	/* read chars until the first newline */
	do{
		c = fgetc(ifp);
		if (c == EOF) {
			printf("\nERROR: Missing new line");
			printf(" (taxon # %d [", notu + 1);
			fputid(stdout, notu);
			printf("], character # %d)!\n", nsite + 1);
			exit(1);
		}
	} while (c != '\n');
}

/* skip control characters and blanks */
void skipcntrl(FILE *ifp, int notu, int nsite)
{
	char c;

	/* read over all control characters and blanks */
	do {
		c = fgetc(ifp);
		if (c == EOF) {
			printf("\nERROR: Missing character ");
			printf(" (taxon # %d [", notu + 1);
			fputid(stdout, notu);
			printf("], character # %d)!\n", nsite + 1);
			exit(1);
		}
	} while (iscntrl(c) || c == ' ');
	/* go one character back */
	if (ungetc(c, ifp) == EOF) {
		printf("\nERROR: File positioning error");
		printf(" (taxon # %d [", notu + 1);
		fputid(stdout, notu);
		printf("], character # %d)!\n", nsite + 1);
		exit(1);
	}
}

/* read sequences of one data set */
void getseqs(FILE *ifp)
{
	int notu, nsite, endofline, linelength, i;
	char c;
	
	seqchars = new_cmatrix(Maxspc, Maxseqc);
	/* read all characters */
	nsite = 0; /* next site to be read */
	while (nsite < Maxseqc) {
		/* read first taxon */
		notu = 0;
		/* go to next true line */
		skiprestofline(ifp, notu, nsite);
		skipcntrl(ifp, notu, nsite);
		if (nsite == 0) readid(ifp, notu);
		endofline = FALSE;
		linelength = 0;		
		do {
			c = readnextcharacter(ifp, notu, nsite + linelength);
			if (c == '\n') endofline = TRUE;
			else if (c == '.') {
				printf(
					"\nERROR: Character '.' not valid in first taxon");
				printf(" ([");
				fputid(stdout, notu);
				printf("], character # %d)!\n",
					nsite + linelength + 1);
				exit(1);
			} else if (nsite + linelength < Maxseqc) {
				/* change to upper case */
				seqchars[notu][nsite + linelength] = toupper(c);
				linelength++;
			} else {
				endofline = TRUE;
				skiprestofline(ifp, notu, nsite + linelength);
			}
		} while (!endofline);	
		if (linelength == 0) {
			printf("\nERROR: Line with length zero ");
			printf(" (taxon # %d [", notu + 1);
			fputid(stdout, notu);
			printf("], character # %d)!\n", nsite + 1);
			exit(1);
		}
		/* read other taxa */
		for (notu = 1; notu < Maxspc; notu++) {
			/* go to next true line */
			if (notu != 1) skiprestofline(ifp, notu, nsite);
			skipcntrl(ifp, notu, nsite);
			if (nsite == 0) readid(ifp, notu);
			for (i = nsite; i < nsite + linelength; i++) {
				c = readnextcharacter(ifp, notu, i);
				if (c == '\n') { /* too short */
					printf("\nERROR: Line too short");
					printf(" (taxon # %d [", notu + 1);
					fputid(stdout, notu);
					printf("], character # %d)!\n", i + 1);
					exit(1);
				} else if (c == '.') {
					seqchars[notu][i] = seqchars[0][i];
				} else {
					/* change to upper case */
					seqchars[notu][i] = toupper(c);
				}
			}
		}
		nsite = nsite + linelength;
	}
}

/* initialize identifer array */
void initid(int t)
{
	int i, j;

	Identif = new_cmatrix(t, 10);
	for (i = 0; i < t; i++)
		for (j = 0; j < 10; j++)
			Identif[i][j] = ' ';
}

/* print identifier of specified taxon in full 10 char length */
void fputid10(FILE *ofp, int t)
{	
	int i;

	for (i = 0; i < 10; i++) fputc(Identif[t][i], ofp);
}

/* print identifier of specified taxon up to first space */
int fputid(FILE *ofp, int t)
{	
	int i;
	
	i = 0;
	while (Identif[t][i] != ' ' && i < 10) {
		fputc(Identif[t][i], ofp);
		i++;
	}
	return i;
}

/* read first line of sequence data set */
void getsizesites(FILE *ifp)
{
	if (fscanf(ifp, "%d", &Maxspc) != 1) {
			printf("\nERROR: Missing number of taxa!\n");
			exit(1);
	}
	if (fscanf(ifp, "%d", &Maxseqc) != 1) {
			printf("\nERROR: Missing number of sites!\n");
			exit(1);
	}
	
	if (Maxspc < 4) {
		printf(
			"\nERROR: Only %d taxa in data set!\n",
				Maxspc);
		exit(1);
	}
	if (Maxspc > 568) {
		printf(
			"\nERROR: Puzzle can do quartet puzzling only for");
		printf(" up to 568 taxa!\n");
		exit(1);
	}
	if (Maxseqc < 1) {
		printf(
			"\nERROR: First line of input file pecifies zero sites!\n");
		exit(1);
	}
	Maxbrnch = 2*Maxspc - 3;
}

/* read one data set - PHYLIP interleaved */
void getdataset(FILE *ifp)
{
	initid(Maxspc);
	getseqs(ifp);
}

/* guess data type */
int guessdatatype()
{
	ulint numnucs, numchars;
	int notu, nsite;
	char c;
	
	/* count A, C, G, T, U, N */
	numnucs = 0;
	numchars = 0;
	for (notu = 0; notu < Maxspc; notu++)
		for (nsite = 0; nsite < Maxseqc; nsite++) {
			c = seqchars[notu][nsite];
			if (c == 'A' || c == 'C' || c == 'G' ||
			    c == 'T' || c == 'U' || c == 'N') numnucs++;
			if (c != '-' && c != '?') numchars++;
		}
	if (numchars == 0) numchars = 1;
	/* more than 85 % frequency means nucleotide data */
	if ((double) numnucs / (double) numchars > 0.85) return 0;
	else return 1;
}

/* translate characters into format used by ML engine */
void translatedataset()
{	
	int notu, sn, co;
	char c;
	cvector code;
	
	codelength = getcodelength(); /* doublets ? */
	Maxsite = Maxseqc / codelength; /* number of ML sites per taxon */
	
	if (Seqchar != NULL) free_cmatrix(Seqchar);
	Seqchar = new_cmatrix(Maxspc, Maxsite);
	
	code = new_cvector(codelength);
	/* decode characters */
	for (notu = 0; notu < Maxspc; notu++) {
		for (sn = 0; sn < Maxsite; sn++) {
			for (co = 0; co < codelength; co++) {
				c = seqchars[notu][sn*codelength + co];
				code[co] = c;
			}
			Seqchar[notu][sn] = code2int(code);
		}
	}
	free_cvector(code);
}

/* estimate mean base frequencies from data set */
void estimatebasefreqs()
{
	int tpmradix, i, j;
	ulint all, *gene;
	
	tpmradix = gettpmradix();
	
	if (Freqtpm != NULL) free_dvector(Freqtpm);
	Freqtpm = new_dvector(tpmradix);
	
	if (Basecomp != NULL) free_imatrix(Basecomp);
	Basecomp = new_imatrix(Maxspc, tpmradix);
	
	gene = (ulint *) malloc((unsigned) ((tpmradix + 1) * sizeof(ulint)));
	if (gene == NULL) maerror("gene [estimatebasefrequencies]");
	
	for (i = 0; i < tpmradix + 1; i++) gene[i] = 0;
	for (i = 0; i < Maxspc; i++)
		for (j = 0; j < tpmradix; j++) Basecomp[i][j] = 0;
	for (i = 0; i < Maxspc; i++)
		for (j = 0; j < Maxsite; j++) {
			gene[(int) Seqchar[i][j]]++;
			if (Seqchar[i][j] != tpmradix) Basecomp[i][(int) Seqchar[i][j]]++;
			}

	all = Maxspc * Maxsite - gene[tpmradix];
	if (all != 0) { /* normal case */
		for (i = 0; i < tpmradix; i++)
			Freqtpm[i] = (double) gene[i] / (double) all;
	} else { /* pathological case with no unique character in data set */
		for (i = 0; i < tpmradix; i++)
			Freqtpm[i] = 1.0 / (double) tpmradix;
	}
	
	free(gene);
	
	Frequ_optn = TRUE;
}

/* guess model of sequence evolution */
void guessmodel()
{
	double c1, c2, c3;
	dvector f;
	dmatrix a;
	int i;

	Dayhf_optn = FALSE;
	Jtt_optn = TRUE;
	mtrev_optn = FALSE;
	AlphaBeta = 4.0;
	AlphaYR = 0.9;
	optim_optn = TRUE;
	HKY_optn = FALSE;
	TN_optn = TRUE;
	numcats = 1;
	Galpha = 0.5;
	lambda = 0.05;
	rate_optim = FALSE;
	lambda_optim = FALSE;

#ifdef SIMULATION
	AlphaBeta = 4.0;
	AlphaYR = 1.0;
	HKY_optn = TRUE;
	TN_optn = FALSE;
#endif
	
	if (data_optn == 1) { /* amino acids */
		
		/* chi2 fit to amino acid frequencies */
		
		f = new_dvector(20);
		a = new_dmatrix(20,20);
		/* chi2 distance Dayhoff */
		dyhfdata(a, f);
		c1 = 0;
		for (i = 0; i < 20; i++)
			c1 = c1 + (Freqtpm[i]-f[i])*(Freqtpm[i]-f[i]);
		/* chi2 distance JTT */
		jttdata(a, f);
		c2 = 0;
		for (i = 0; i < 20; i++)
			c2 = c2 + (Freqtpm[i]-f[i])*(Freqtpm[i]-f[i]);
		/* chi2 distance mtREV */
		mtrevdata(a, f);
		c3 = 0;
		for (i = 0; i < 20; i++)
			c3 = c3 + (Freqtpm[i]-f[i])*(Freqtpm[i]-f[i]);
		free_dvector(f);
		free_dmatrix(a);
		if (c1 < c2) { /* Dayhoff or mtREV */
			if (c3 < c1) { /* mtREV */
				Dayhf_optn = FALSE;
				Jtt_optn = FALSE;
				mtrev_optn = TRUE;
				printf(
					"\nINFO:  Assuming amino acids encoded on mtDNA\n");
			} else { /* Dayhoff */
				Dayhf_optn = TRUE;
				Jtt_optn = FALSE;
				mtrev_optn = FALSE;
				printf(
					"\nINFO:  Assuming amino acids encoded on nuclear DNA\n");
			}
		} else { /* JTT or mtREV */
			if (c3 < c2) { /* mtREV */
				Dayhf_optn = FALSE;
				Jtt_optn = FALSE;
				mtrev_optn = TRUE;
				printf(
					"\nINFO:  Assuming amino acids encoded on mtDNA\n");
			} else { /* JTT */
				Dayhf_optn = FALSE;
				Jtt_optn = TRUE;
				mtrev_optn = FALSE;
				printf(
					"\nINFO:  Assuming amino acids encoded on nuclear DNA\n");
			}
		}
	} else { /* nucleotides */
		printf("\nINFO:  Assuming Tamura-Nei (1993) model of sequence evolution\n");
	}
}


/******************************************************************************/
/* options - file I/O - output                                                */
/******************************************************************************/

/* compute TN parameters according to F84 Ts/Tv ratio */
void makeF84model()
{
	double rho, piA, piC, piG, piT, piR, piY, ab, yr;
	
	piA = Freqtpm[0];
	piC = Freqtpm[1];
	piG = Freqtpm[2];
	piT = Freqtpm[3];
	piR = piA + piG;
	piY = piC + piT;
	if (piC*piT*piR + piA*piG*piY == 0.0) {		
		printf("\n\nERROR:  F84 model not possible ");
		printf("(bad combination of base frequencies)\n");
		tstvf84 = 0.0;
		return;
	}
	rho = (piR*piY*(piR*piY*tstvf84 - (piA*piG + piC*piT)))/
		(piC*piT*piR + piA*piG*piY);
	
	if(piR == 0.0 || piY == 0.0 || (piR + rho) == 0.0) {
		printf("\n\nERROR:  F84 model not possible ");
		printf("(bad combination of base frequencies)\n");
		tstvf84 = 0.0;
		return;
	}
	ab = 1 + 0.5*rho*(1.0/piR + 1.0/piY);
	yr = (piY + rho)/piY * piR/(piR + rho);
	if (ab < MINAB || ab > MAXAB) {
		printf("\n\nERROR:  F84 model not possible ");
		printf("(bad Ts/Tv parameter)\n");
		tstvf84 = 0.0;
		return;
	}
	if (yr < MINYR || yr > MAXYR) {
		printf("\n\nERROR:  F84 model not possible ");
		printf("(bad Y/R transition parameter)\n");
		tstvf84 = 0.0;
		return;
	}
	AlphaBeta = ab;
	AlphaYR = yr;
	optim_optn = FALSE;
}

/* set options interactively */
void setoptions()
{	
	int ch, i;
	double sumfreq;

	/* defaults */
	Numtrial = 1000;    /* number of puzzling steps       */
	outgroup = 0;       /* use first taxon as outgroup    */
	sym_optn = FALSE;   /* symmetrize doublet frequencies */
	tstvf84 = 0.0;      /* disable F84 model              */
	tri_optn = FALSE;   /* no characteristic triangle     */
	show_optn = FALSE;  /* show unresolved quartets       */

 	do {	
		printf("\n\nSettings for this analysis:\n\n",
			VERSION);
		printf("k                          Tree search?  ");
		if (puzzlemode == 0) printf("Quartet puzzling\n");
		if (puzzlemode == 1) printf("User defined tree\n");
		if (puzzlemode == 2) printf("No tree (ML distances only)\n");
		if (puzzlemode == 0) {
			printf("n             Number of puzzling steps?  %lu\n",
					Numtrial);
			printf("u             Show unresolved quartets?  %s\n",
				(show_optn ? "Yes" : "No"));
		/* NOT YET */
		/*	printf("g     Generate characteristic triangle?  %s\n",
				(tri_optn ? "Yes" : "No")); */
			printf("o                    Selected outgroup?  ");
			fputid(stdout, outgroup);
			printf("\n");
		}
		printf("d          Type of sequence input data?  %s\n",
			(data_optn ? "Amino acids" : "Nucleotides"));		
		printf("m          Model of sequence evolution?  ");
		
		if (data_optn == 0) { /* nucleotides */
			if (nuc_optn) {
				if(HKY_optn)
					printf("HKY (Hasegawa et al. 1985)\n");	
				else {
					printf("TN (Tamura-Nei 1993)\n");
					printf("p      Constrain TN model to F84 model?  ");
					if (tstvf84 == 0.0)
						printf("No\n");
					else printf("Yes (Ts/Tv ratio = %.2f)\n", tstvf84);
				}
				printf(
					"t    Transition/transversion parameter?  ");
				if (optim_optn)
					printf("Estimate from data set\n");
				else
					printf("%.2f\n", AlphaBeta/2.0);	
				if (TN_optn) {
					printf(
						"r             Y/R transition parameter?  ");
					if (optim_optn)
						printf("Estimate from data set\n");
					else
						printf("%.2f\n", AlphaYR);		
 				}			
			}
			if (SH_optn) {
					printf("SH (Schoeniger-von Haeseler 1994)\n");
					printf(
						"t    Transition/transversion parameter?  ");
					if (optim_optn)
						printf("Estimate from data set\n");
					else
						printf("%.2f\n", AlphaBeta/2.0);
			}	
		}
		if (data_optn == 1) { /* amino acids */
			if (Dayhf_optn) printf("Dayhoff (Dayhoff et al. 1978)\n");	
			if (Jtt_optn) printf("JTT (Jones et al. 1992)\n");
			if (mtrev_optn) printf("mtREV-22 (Adachi-Hasegawa 1995)\n");
		}
		if (data_optn == 1)
			printf("f               Amino acid frequencies?  ");
		else if (data_optn == 0 && SH_optn)
			printf("f                  Doublet frequencies?  ");
		else if (data_optn == 0 && nuc_optn)
			printf("f               Nucleotide frequencies?  ");
		printf("%s\n", (Frequ_optn ? "Estimate from data set" :
			"Use specified values"));
		if (data_optn == 0 && SH_optn)
			printf("s       Symmetrize doublet frequencies?  %s\n",
				(sym_optn ? "Yes" : "No"));
		
		/* NOT YET */
		/* printf("c            Number of rate categories?  %d\n", numcats);
		if (numcats > 1) {
			printf("a          Rate heterogenity parameter?  ");
			if (rate_optim)
				printf("Estimate from data set\n");
			else if (Galpha < 1.0)
				printf("%.2f (strong rate heterogenity)\n", Galpha);
			else printf("%.2f (weak rate heterogenity)\n", Galpha);
			
			printf("l           Rate correlation parameter?  ");
			if (lambda_optim)
				printf("Estimate from data set\n");
			else if (lambda == 0.0)
				printf("0.00 (no correlation)\n");
			else
				printf("%.2f\n", lambda);
		} */
		
		printf("\nAre these settings OK ");
		printf("(type y or choose from the menu): ");
		
		/* read one char */
		ch = getchar();
		if (ch != '\n') {
			do ;
			while (getchar() != '\n');
		}
		ch = tolower(ch);
		
		/* letters in use: a c d f g k l m n o p r s t u y */

		switch (ch) {

			case '\n':	break;
			
			case 'k':	puzzlemode = puzzlemode + 1;
						if (puzzlemode == 3) puzzlemode = 0;
						break;

			case 'd':	data_optn = 1 - data_optn;
						/* translate characters into format used by ML engine */
	                    translatedataset();
	                    estimatebasefreqs();
						break;

			/* NOT YET */
			/* case 'g':	if (puzzlemode == 0)
							tri_optn = 1 - tri_optn;
						else
							printf("\n\nThis is not a possible option!\n");
						break; */

			case 'u':	if (puzzlemode == 0)
							show_optn = 1 - show_optn;
						else
							printf("\n\nThis is not a possible option!\n");
						break;

			case 'f':	if (Frequ_optn) {
							tstvf84 = 0.0;
							Frequ_optn = FALSE;
							sumfreq = 0.0;
							if (data_optn == 1)
								printf("\n\nAmino acid");
							else if (data_optn == 0 && SH_optn)
								printf("\n\nDoublet");
							else if (data_optn == 0 && nuc_optn)
								printf("\n\nNucleotide");
							printf(" frequencies (in percent):\n\n");
							for (i = 0; i < gettpmradix() - 1; i++) {
								printf("pi(%s) = ", int2code(i));
								scanf("%lf", &(Freqtpm[i]));
								do ;
								while (getchar() != '\n');
								Freqtpm[i] = Freqtpm[i]/100.0;
								if (Freqtpm[i] < 0.0) {
									printf(
										"\n\nERROR: Negative frequency!\n");
									estimatebasefreqs();
									break;
								}
								sumfreq = sumfreq + Freqtpm[i];
								if (sumfreq > 1.0) {
									printf("\n\nERROR: The sum of ");
									printf("all frequencies exceeds");
									printf(" 100%%!\n");
									estimatebasefreqs();
									break;
								}
								if (i == gettpmradix() - 2)
									Freqtpm[i+1] = 1.0 - sumfreq;
							}
						} else estimatebasefreqs();
						break;
			
			case 's':	if (data_optn == 0 && SH_optn) {
							sym_optn = 1 - sym_optn;
						} else {
							printf(
								"\n\nThis is not a possible option!\n");
						}
						break;

			case 'n':	if (puzzlemode == 0) {
							printf(
								"\n\nNumber of puzzling steps: ");
							scanf("%lu", &Numtrial);
							do ;
							while (getchar() != '\n');
							if (Numtrial < 1) {
								printf(
									"\n\nERROR: The number of puzzling");
								printf(
									" steps can't be smaller than one!\n");
								Numtrial = 100;
							}
						} else {
							printf("\n\nThis is not a possible option!\n");
						}
						break;
						
			case 'o':	if (puzzlemode == 0) {
							printf(
								"\n\nTaxon number of the outgroup (1-%d): ",
									Maxspc);
							scanf("%d", &outgroup);
							do ;
							while (getchar() != '\n');
							if (outgroup < 1 || outgroup > Maxspc) {
								printf("\n\nERROR: Taxon numbers ");
								printf("range from 1 to %d only!\n",
										Maxspc);
								outgroup = 1;
							}
							outgroup = outgroup - 1;
						} else {
							printf("\n\nThis is not a possible option!\n");
						}
						break;
	
			case 'm':	if (data_optn == 0) { /* nucleotide data */
							if(HKY_optn && nuc_optn) {
								/* HKY -> TN */
								tstvf84 = 0.0;
								AlphaBeta = 4.0;
								AlphaYR = 0.9;
								HKY_optn = FALSE;
								TN_optn = TRUE;
								optim_optn = TRUE;
								nuc_optn = TRUE;
								SH_optn = FALSE;
								break;
							}
							if(TN_optn && nuc_optn) {
								if (Maxseqc % 2 == 0) {
									/* number of chars is a multiple 2 */
									/* (doublets) */
									/* TN -> SH */
									tstvf84 = 0.0;
									AlphaBeta = 4.0;
									AlphaYR = 1.0;
									HKY_optn = TRUE;
									TN_optn = FALSE;
									optim_optn = TRUE;
									nuc_optn = FALSE;
									SH_optn = TRUE;
									/* translate characters into format */
									/* used by ML engine */
									translatedataset();
									estimatebasefreqs();
								} else {
									printf("\n\nINFO:  SH model not ");
									printf("available (odd number of");
									printf(" sites)\n");
									/* TN -> HKY */
									tstvf84 = 0.0;
									AlphaBeta = 4.0;
									AlphaYR = 1.0;
									HKY_optn = TRUE;
									TN_optn = FALSE;
									optim_optn = TRUE;
									nuc_optn = TRUE;
									SH_optn = FALSE;
								}
								break;
							}
							if(SH_optn) {
								/* SH -> HKY */
								tstvf84 = 0.0;
								AlphaBeta = 4.0;
								AlphaYR = 1.0;
								HKY_optn = TRUE;
								TN_optn = FALSE;
								optim_optn = TRUE;
								nuc_optn = TRUE;
								SH_optn = FALSE;
								/* translate characters into format */
								/* used by ML engine */
								translatedataset();
								estimatebasefreqs();
								break;
							}
							break;
						}
						if (data_optn == 1) { /* amino acid data */
							if (Dayhf_optn) {
								/* Dayhoff -> JTT */
								Dayhf_optn = FALSE;
								Jtt_optn = TRUE;
								mtrev_optn = FALSE;
								break;
							}
							if (Jtt_optn) {
								/* JTT -> mtREV */
								Dayhf_optn = FALSE;
								Jtt_optn = FALSE;
								mtrev_optn = TRUE;
								break;
							}
							if (mtrev_optn) {
								/* mtREV -> Dayhoff */
								Dayhf_optn = TRUE;
								Jtt_optn = FALSE;
								mtrev_optn = FALSE;
								break;
							}
							break;
						}
						break;
						
			case 't':	if (data_optn == 1) {
							printf(
								"\n\nThis is not a possible option!\n");
						} else {
							tstvf84 = 0.0;
							printf(
								"\n\nEnter an invalid value for ");
							printf(
								"estimation from data set!\n");
							printf(
								"\nTransition/transversion parameter (%.2f-%.2f): ",
									0.5*MINAB, 0.5*MAXAB);
							scanf("%lf", &AlphaBeta);
							do ;
							while (getchar() != '\n');
							AlphaBeta = 2 * AlphaBeta;
							if (AlphaBeta < MINAB || AlphaBeta > MAXAB) {
								optim_optn = TRUE;
								AlphaBeta = 4.0;
							} else {
								optim_optn = FALSE;
							}
						}
						break;

			case 'r':	if (!(TN_optn && nuc_optn)){
							printf(
								"\n\nThis is not a possible option!\n");
						} else {
							tstvf84 = 0.0;
							printf(
								"\n\nThis parameter corresponds to ");
							printf(
								"alpha2/alpha1 (Tamura-Nei 1993).");
							printf(
								"\nEnter an invalid value ");
							printf(
								"for estimation from data set!\n");
							printf("\nY/R transition parameter (%.2f-%.2f): ", MINYR, MAXYR);
							scanf("%lf", &AlphaYR);
							do ;
							while (getchar() != '\n');
							if (AlphaYR < MINYR || AlphaYR > MAXYR) {
								optim_optn = TRUE;
								AlphaYR = 0.9;
							} else if (AlphaYR == 1.0) {
								TN_optn = FALSE;
								HKY_optn = TRUE;
								if (optim_optn) AlphaBeta = 4.0;
							} else {
								optim_optn = FALSE;
							}
						}
						break;
						
			case 'p':	if (!(TN_optn && nuc_optn)){
							printf(
								"\n\nThis is not a possible option!\n");
						} else {
							printf(
								"\n\nThe F84 model (Felsenstein 1984) is a restricted");
							printf(
								" TN model, and the one\nF84 parameter uniquely");
							printf(
								" determines the two corresponding TN parameters!\n\n");
							printf(
								"F84 expected transition/transversion ratio: ");
							scanf("%lf", &tstvf84);
							do ;
							while (getchar() != '\n');
							if (tstvf84 <= 0.0) tstvf84 = 0.0;
							else makeF84model();
						}
						break;

			/* NOT YET */
			/* case 'a':	if (numcats > 1) {
							printf(
								"\n\nEnter an invalid ");
							printf(
								"for estimation from data set!\n");
							printf(
								"\nRate heterogenity parameter (%.2f-%.2f): ", MINGA, MAXGA);
							scanf("%lf", &Galpha);
							do ;
							while (getchar() != '\n');
							if (Galpha < MINGA || Galpha > MAXGA) {
								rate_optim = TRUE;
								Galpha = 0.5;
							} else {
								rate_optim = FALSE;
							}
						} else 
							printf("\n\nThis is not a possible option!\n");
						break;

			case 'l':	if (numcats > 1) {
							printf(
								"\n\nEnter an invalid value ");
							printf(
								"for estimation from data set!\n");
							printf(
								"\nRate correlation parameter (%.2f-%.2f): ", MINLA, MAXLA);
							scanf("%lf", &lambda);
							do ;
							while (getchar() != '\n');
							if (lambda < MINLA || lambda > MAXLA) {
								lambda_optim = TRUE;
								lambda = 0.2;
							} else {
								lambda_optim = FALSE;
							}
							if (lambda == 1.0) {
								rate_optim = FALSE;
								lambda_optim = FALSE;
								numcats = 1;
								lambda = 0.0;
								Galpha = 0.5;
							}						
						} else 
							printf("\n\nThis is not a possible option!\n");
						break;
						
			case 'c':	printf(
							"\n\nNumber of rate categories (%d-%d, 4 are recommended): ", MINCAT, MAXCAT);
						scanf("%d", &numcats);
						do ;
						while (getchar() != '\n');
						if (numcats < MINCAT)
							numcats = MINCAT;
						if (numcats > MAXCAT)
							numcats = MAXCAT;
						if (numcats == 1) {
							rate_optim = FALSE;
							lambda_optim = FALSE;
							Galpha = 0.5;
							lambda = 0.05;
						} else {
							rate_optim = FALSE;
							lambda_optim = FALSE;
							Galpha = 0.5;
							lambda = 0.0;
						}
						break;
			*/

			case 'y':	break;

			default:	printf("\n\nThis is not a possible option!\n");
						break;
		}
	} while (ch != 'y');

	printf("\n");
}

/* open file for reading */
void openfiletoread(FILE **fp, char name[])
{
	char inputline[64];

	if ((*fp = fopen(name, "r")) == NULL) {
   		printf(
   			"\nERROR: Missing file '%s'!\n",
   				name);
		printf(
			"       Please enter another name for this file: ");
		gets(inputline);
		if ((*fp = fopen(inputline, "r")) == NULL) {
			printf(
				"\nERROR: Missing file '%s'!\n",
					inputline);
			exit(1);
		}
	}
}

/* open file for writing */
void openfiletowrite(FILE **fp, char name[])
{
	char inputline[64];

	if ((*fp = fopen(name, "w")) == NULL) {
   		printf(
   			"\nERROR: Can't create file '%s'!\n",
   				name);
		printf(
			"       Please enter another name for this file: ");
		gets(inputline);
		if ((*fp = fopen(inputline, "w")) == NULL) {
			printf(
				"\nERROR: Can't create file '%s'!\n",
					inputline);
			exit(1);
		}
	}
}

/* close file */
void closefile(FILE *fp)
{	
	fclose(fp);
}

/* show ML base frequencies */
void showfrequencies()
{	
	int i;

	if (data_optn == 1)
			printf("\nINFO:  Amino acid ");
		else if (data_optn == 0 && SH_optn)
			printf("\nINFO:  Doublet ");
		else if (data_optn == 0 && nuc_optn)
			printf("\nINFO:  Nucleotide ");
	printf("frequencies used in maximum likelihood calculation:\n\n");
				
	for (i = 0; i < gettpmradix(); i++)
		printf("        pi(%s) = %.1f%%\n",
			int2code(i), Freqtpm[i]*100);
}

/* symmetrize doublet frequencies */
void symdoublets()
{
	int i, imean;
	double mean;
	
	if (data_optn == 0 && SH_optn && sym_optn) {
		/* ML frequencies */
		mean = (Freqtpm[1] + Freqtpm[4])/2.0; /* AC CA */
		Freqtpm[1] = mean;
		Freqtpm[4] = mean;
		mean = (Freqtpm[2] + Freqtpm[8])/2.0; /* AG GA */
		Freqtpm[2] = mean;
		Freqtpm[8] = mean;
		mean = (Freqtpm[3] + Freqtpm[12])/2.0; /* AT TA */
		Freqtpm[3] = mean;
		Freqtpm[12] = mean;
		mean = (Freqtpm[6] + Freqtpm[9])/2.0; /* CG GC */
		Freqtpm[6] = mean;
		Freqtpm[9] = mean;
		mean = (Freqtpm[7] + Freqtpm[13])/2.0; /* CT TC */
		Freqtpm[7] = mean;
		Freqtpm[13] = mean;
		mean = (Freqtpm[11] + Freqtpm[14])/2.0; /* GT TG */
		Freqtpm[11] = mean;
		Freqtpm[14] = mean;
		
		/* base composition of each taxon */
		for (i = 0; i < Maxspc; i++) {
			imean = (Basecomp[i][1] + Basecomp[i][4])/2; /* AC CA */
			Basecomp[i][1] = imean;
			Basecomp[i][4] = imean;
			imean = (Basecomp[i][2] + Basecomp[i][8])/2; /* AG GA */
			Basecomp[i][2] = imean;
			Basecomp[i][8] = imean;
			imean = (Basecomp[i][3] + Basecomp[i][12])/2; /* AT TA */
			Basecomp[i][3] = imean;
			Basecomp[i][12] = imean;
			imean = (Basecomp[i][6] + Basecomp[i][9])/2; /* CG GC */
			Basecomp[i][6] = imean;
			Basecomp[i][9] = imean;
			imean = (Basecomp[i][7] + Basecomp[i][13])/2; /* CT TC */
			Basecomp[i][7] = imean;
			Basecomp[i][13] = imean;
			imean = (Basecomp[i][11] + Basecomp[i][14])/2; /* GT TG */
			Basecomp[i][11] = imean;
			Basecomp[i][14] = imean;
		}
	}
}

/* show Ts/Tv ratio and Ts Y/R ratio */
void shownucratios()
{
	double AlphaYBeta, AlphaRBeta, piR, piY, num, denom, pyr, pur;
	
	if (nuc_optn == TRUE) { /* 4x4 nucs */
		piR = Freqtpm[0] + Freqtpm[2];
		piY = Freqtpm[1] + Freqtpm[3];
		AlphaRBeta = 2 * AlphaBeta / (1 + AlphaYR);
		AlphaYBeta = AlphaRBeta * AlphaYR;
		tstvratio = (AlphaRBeta*Freqtpm[0]*Freqtpm[2] +
					 AlphaYBeta*Freqtpm[1]*Freqtpm[3])/(piR * piY);
		yrtsratio = (AlphaYBeta*Freqtpm[1]*Freqtpm[3]) /
					(AlphaRBeta*Freqtpm[0]*Freqtpm[2]);
	} else { /* 16x16 nucs */
		pyr = Freqtpm[1]*Freqtpm[3] + Freqtpm[5]*Freqtpm[7] +
			Freqtpm[9]*Freqtpm[11] + Freqtpm[4]*Freqtpm[12] +
			Freqtpm[5]*Freqtpm[13] + Freqtpm[6]*Freqtpm[14] +
			Freqtpm[7]*Freqtpm[15] + Freqtpm[13]*Freqtpm[15];
		pur = Freqtpm[0]*Freqtpm[2] + Freqtpm[4]*Freqtpm[6] +
			Freqtpm[0]*Freqtpm[8] + Freqtpm[1]*Freqtpm[9] +
			Freqtpm[2]*Freqtpm[10] + Freqtpm[8]*Freqtpm[10] +
			Freqtpm[3]*Freqtpm[11] + Freqtpm[12]*Freqtpm[14];
		num = pyr + pur;
		denom = Freqtpm[0]*Freqtpm[1] + Freqtpm[1]*Freqtpm[2] +
			Freqtpm[0]*Freqtpm[3] + Freqtpm[2]*Freqtpm[3] +
			Freqtpm[0]*Freqtpm[4] + Freqtpm[1]*Freqtpm[5] +
			Freqtpm[4]*Freqtpm[5] + Freqtpm[2]*Freqtpm[6] +
			Freqtpm[5]*Freqtpm[6] + Freqtpm[3]*Freqtpm[7] +
			Freqtpm[4]*Freqtpm[7] + Freqtpm[6]*Freqtpm[7] +
			Freqtpm[4]*Freqtpm[8] + Freqtpm[5]*Freqtpm[9] +
			Freqtpm[8]*Freqtpm[9] + Freqtpm[6]*Freqtpm[10] +
			Freqtpm[9]*Freqtpm[10] + Freqtpm[7]*Freqtpm[11] +
			Freqtpm[8]*Freqtpm[11] + Freqtpm[10]*Freqtpm[11] +
			Freqtpm[0]*Freqtpm[12] + Freqtpm[8]*Freqtpm[12] +
			Freqtpm[1]*Freqtpm[13] + Freqtpm[9]*Freqtpm[13] +
			Freqtpm[12]*Freqtpm[13] + Freqtpm[2]*Freqtpm[14] +
			Freqtpm[10]*Freqtpm[14] + Freqtpm[13]*Freqtpm[14] +
			Freqtpm[3]*Freqtpm[15] + Freqtpm[11]*Freqtpm[15] +
			Freqtpm[12]*Freqtpm[15] + Freqtpm[14]*Freqtpm[15];
		tstvratio = AlphaBeta * num/denom;
		yrtsratio = pyr/pur;
	}
	printf(
		"\nINFO:  Expected transition/transversion ratio: %.2f\n", tstvratio);
	printf("\nINFO:  Expected pyrimidine transition/purine");
	printf(" transition ratio: %.2f\n", yrtsratio);

}

/* show rate categories */
void showrates()
{
	int i;

	printf("\nINFO:  Relative rates of the %d", numcats);
	printf(" equally probable rate categories:\n\n");
				
	for (i = 0; i < numcats; i++)
		printf("        Category %d: %.4f\n",
			i+1, Rates[i]);
}

/* show average rate patch length */
void showpatchlength(FILE *fp)
{
	double sum;
	
	sum = 1.0/ (double)numcats;
	sum = (1.0 - sum)*(1.0 - lambda);
	fprintf(fp, "Average rate patch length: ");
	if (sum == 0.0) fprintf(fp, "Infinite");
	else fprintf(fp, "%.2f", 1.0/sum);
}

/* write ML distance matrix to file */
void putdistance(FILE *fp)
{
	int i, j;
	
	fprintf(fp, "  %d\n", Maxspc);
	for (i = 0; i < Maxspc; i++) {
		fputid10(fp, i);
		for (j = 0; j < Maxspc; j++) {
			fprintf(fp, "  %.4f", Distanmat[i][j]/100.0);
			/* eight in one row */
			if ((j + 1) % 8 == 0 && j+1 != Maxspc)
				fprintf(fp, "\n          ");
		}
		fprintf(fp, "\n");
	}
}

void writeoutputfile(FILE *ofp)
{
	int i, fail;
	ulint li;
	double pval;

	fprintf(ofp, "PUZZLE (version %s)\n\n", VERSION);
	fprintf(ofp, "Session of %s", asctime(localtime(&Starttime)) );
	fprintf(ofp, "Runtime: %lu seconds (= %.1f minutes = %.1f hours)\n\n\n",
		Stoptime - Starttime, (Stoptime - Starttime)/60.,
		(Stoptime - Starttime)/3600.);
	fprintf(ofp, "MAXIMUM LIKELIHOOD MODEL\n\n");
	fprintf(ofp, "Sequence input data: %d taxa with %d ", Maxspc, Maxsite);
	if (data_optn == 1)
		fprintf(ofp, "amino acid");
	else if (data_optn == 0 && SH_optn)
		fprintf(ofp, "doublet (%d nucleotide)", Maxseqc);
	else if (data_optn == 0 && nuc_optn)
		fprintf(ofp, "nucleotide");
	fprintf(ofp, " sites\n");
	fprintf(ofp, "Model of sequence evolution: ");
	if (data_optn == 0) { /* nucleotides */
		if (nuc_optn) {
			if(HKY_optn) fprintf(ofp, "HKY (Hasegawa et al. 1985)\n");	
			else fprintf(ofp, "TN (Tamura-Nei 1993)\n");
			fprintf(ofp, "Transition/transversion parameter");
			if (optim_optn)
				fprintf(ofp, " (estimated from data set)");
			fprintf(ofp, ": %.2f", AlphaBeta/2.0);
			if (optim_optn)
				fprintf(ofp, " +/- %.2f", tserr);
			fprintf(ofp, "\n");	
			if (TN_optn) {
				fprintf(ofp, "Y/R transition parameter");
				if (optim_optn)
					fprintf(ofp, " (estimated from data set)");
				fprintf(ofp, ": %.2f", AlphaYR);
				if (optim_optn)
					fprintf(ofp, " +/- %.2f", yrerr);
				fprintf(ofp, "\n");	
			}
		}
		if (SH_optn) {
			fprintf(ofp, "SH (Schoeniger-von Haeseler 1994)\n");
			fprintf(ofp, "Transition/transversion parameter");
			if (optim_optn) fprintf(ofp, " (estimated from data set)");
			fprintf(ofp, ": %.2f\n", AlphaBeta/2.0);
		}	
	}
	if (data_optn == 1) { /* amino acids */
		if (Dayhf_optn) fprintf(ofp, "Dayhoff (Dayhoff et al. 1978)\n");	
		if (Jtt_optn) fprintf(ofp, "JTT (Jones et al. 1992)\n");
		if (mtrev_optn) fprintf(ofp, "mtREV-22 (Adachi-Hasegawa 1995)\n");
	}
	if (data_optn == 1)
			fprintf(ofp, "Amino acid ");
		else if (data_optn == 0 && SH_optn)
			fprintf(ofp, "Doublet ");
		else if (data_optn == 0 && nuc_optn)
			fprintf(ofp, "Nucleotide ");
	fprintf(ofp, "frequencies (");
	if (Frequ_optn) fprintf(ofp, "estimated from data set");
	else fprintf(ofp, "user specified");
	if (data_optn == 0 && SH_optn && sym_optn)
		fprintf(ofp, " and symmetrized");
	fprintf(ofp, "):\n\n");
	for (i = 0; i < gettpmradix(); i++)
		fprintf(ofp, " pi(%s) = %.1f%%\n",
			int2code(i), Freqtpm[i]*100);
	if (data_optn == 0) {
		fprintf(ofp, "\nExpected transition/transversion ratio: %.2f",
			tstvratio);
		if (tstvf84 == 0.0) fprintf(ofp, "\n");
		else fprintf(ofp, " (= F84 parameter)\n");
		fprintf(ofp, "Expected pyrimidine transition/purine transition");
		fprintf(ofp, " ratio: %.2f\n", yrtsratio);
		if (tstvf84 != 0.0 && TN_optn)
			fprintf(ofp,
				"This TN model is equivalent to a F84 model (Felsenstein 1984)\n");
	}
	
	/* NOT YET */
	/* fprintf(ofp, "\n\nRATE VARIATION\n\n");	
	if (numcats != 1) {
		fprintf(ofp, "Heterogenity parameter");
		if (rate_optim) fprintf(ofp, " (estimated from data set)");
		fprintf(ofp, ": %.2f ", Galpha);
		if (rate_optim) fprintf(ofp, "+/- %.2f ", gaerr);	
		if (Galpha < 1.0) fprintf(ofp, "(strong rate heterogenity)\n");
		else fprintf(ofp, "(weak rate heterogenity)\n");
		fprintf(ofp, "Correlation parameter");
		if (lambda_optim) fprintf(ofp, " (estimated from data set)");
		fprintf(ofp, ": %.2f ", lambda);
		if (lambda_optim) fprintf(ofp, "+/- %.2f ", laerr);	
		fprintf(ofp, "\n");
		fprintf(ofp, "Relative rates of the %d", numcats);
		fprintf(ofp, " equally probable rate categories:\n\n");				
		for (i = 0; i < numcats; i++)
			fprintf(ofp, " Category %d: %.4f\n",
				i+1, Rates[i]);
		fprintf(ofp, "\nThis discrete distribution approximates a ");
		fprintf(ofp, "Gamma-distribution with\nthe given parameter ");
		fprintf(ofp, "(expectation 1, variance 1/%.2f = %.2f)\n\n", Galpha,
			1.0/Galpha);
		showpatchlength(ofp);
		fprintf(ofp, "\n");
	} else {
		fprintf(ofp, "Only one rate category assumed");
		fprintf(ofp, " (no rate variation)\n");
	} */

	fprintf(ofp, "\n\nTREE SEARCH\n\n");

	if (puzzlemode == 0) {
		fprintf(ofp, "Quartet puzzling is used to choose from the possible tree topologies\n");
		fprintf(ofp, "and to simultaneously infer reliability values for internal branches\n\n");
		fprintf(ofp, "Number of puzzling steps: %lu\n", Numtrial);
		fprintf(ofp, "Analysed quartets: %lu\n", Numquartets);
		fprintf(ofp, "Unresolved quartets: %lu (= %.1f%%)\n",
			badqs, (double) badqs / (double) Numquartets * 100);
		fprintf(ofp, "Quartet trees are based on approximate ");
		fprintf(ofp, "maximum likelihood values\n\n\n");
	}
	if (puzzlemode == 1) {
		fprintf(ofp, "The tree topology is specified in the input tree file\n");
	}
	if (puzzlemode == 2) {
		fprintf(ofp, "No tree search performed (maximum likelihood distances only)\n");
	}
	if (puzzlemode == 0) {
		fprintf(ofp, "QUARTET PUZZLING TREE\n\n");
		fprintf(ofp, "* This is an unrooted tree\n");
		fprintf(ofp,
			"* Support for each internal branch is shown in percent\n");
		fprintf(ofp, "* The tree may be multifurcating\n\n\n");
		plotconsensustree(ofp);
		fprintf(ofp, "\n\nQuartet puzzling tree (in CLUSTAL W notation):\n\n");
		writeconsensustree(ofp);
		fprintf(ofp, "\n\nBIPARTITIONS\n\n");
		fprintf(ofp, "The following bipartitions occured at least once");
		fprintf(ofp, " in all intermediate\ntrees that have been generated ");
		fprintf(ofp, "in the %lu puzzling steps:\n\n", Numtrial);
		fprintf(ofp, "Bipartitions included in the consensus tree:\n");
		fprintf(ofp,
			"(bipartition with taxa in input order : number of times seen)\n\n");
		for (li = 0; li < consincluded; li++) {
			fprintf(ofp, " ");
			printsplit(ofp, splitfreqs[2*li+1]);
			fprintf(ofp, "  :  %lu\n", splitfreqs[2*li]);
		}
		if (consincluded == 0) fprintf(ofp, " None (no bipartition included)\n");
		fprintf(ofp, "\nBipartitions not included in the consensus tree:\n");
		fprintf(ofp,
			"(bipartition with taxa in input order : number of times seen)\n\n");
		for (li = consincluded; li < numbiparts; li++) {
			fprintf(ofp, " ");
			printsplit(ofp, splitfreqs[2*li+1]);
			fprintf(ofp, "  :  %lu\n", splitfreqs[2*li]);
		}
		if (consincluded == numbiparts)
			fprintf(ofp, " None (all bipartitions are included)\n");		
	}

	fprintf(ofp, "\n\nTAXA IN INPUT ORDER\n\n");
	fail = FALSE;
	for (i = 0; i < Maxspc; i++) {
		fprintf(ofp, " ");
		fputid10(ofp, i);
		pval = homogentest(i);
		if ( pval < 0.05 )
			fprintf(ofp, "  * critical significance level: %.2f%%",
				pval*100.0);
		if (chi2fail) fail = TRUE;
		fprintf(ofp, "\n");	
	}
	fprintf(ofp, "\nAll taxa marked * differ on the basis of a ");
	fprintf(ofp, "5%% level chi-square test\nin ");
	if (data_optn == 1)
		fprintf(ofp, "amino acid");
	else if (data_optn == 0 && SH_optn)
		fprintf(ofp, "doublet");
	else if (data_optn == 0 && nuc_optn)
		fprintf(ofp, "nucleotide");
	fprintf(ofp,
		" composition from the frequency distribution of\nthe maximum likelihood model\n");
	if (fail) {
		fprintf(ofp,
			"\nWarning: Result of chi-square test may be not valid");
		fprintf(ofp,
			" because of small\nmaximum likelihood frequencies and");
		fprintf(ofp,
			" short sequence length!\n");
	}

#ifndef SIMULATION
	if (puzzlemode == 0) {
		fprintf(ofp, "\n\nMAXIMUM LIKELIHOOD BRANCH LENGTHS ON QUARTET");
		fprintf(ofp, " PUZZLING TREE\n\nBranch lengths are computed using");
		fprintf(ofp, " the selected model of\nsequence evolution\n\n\n");
		prtopology(ofp);
		fprintf(ofp, "\n");
		resulttree(ofp);
		fprintf(ofp, "\n\nQuartet puzzling tree with maximum likelihood branch lengths");
		fprintf(ofp, "\n(in CLUSTAL W notation):\n\n");
		fputphylogeny(ofp);
	}
	
	if (puzzlemode == 1) {
		fprintf(ofp, "\n\nMAXIMUM LIKELIHOOD BRANCH LENGTHS ON USER");
		fprintf(ofp, " DEFINED TREE\n\nBranch lengths are computed using");
		fprintf(ofp, " the selected model of\nsequence evolution\n\n\n");
		prtopology(ofp);
		fprintf(ofp, "\n");
		resulttree(ofp);
		fprintf(ofp, "\n\nUser defined tree with maximum likelihood branch lengths");
		fprintf(ofp, "\n(in CLUSTAL W notation):\n\n");
		fputphylogeny(ofp);
	}
#endif
	
	fprintf(ofp, "\n\nMAXIMUM LIKELIHOOD DISTANCES\n\n");
	fprintf(ofp, "Maximum likelihood distances are computed using the ");
	fprintf(ofp, "selected model of\nsequence evolution\n\n");
	putdistance(ofp);

}


/******************************************************************************/
/* main part                                                                  */
/******************************************************************************/

main()
{	
	int i, j, a, b, c, k, ID;
	ulint nqused;
	ivector freeID;
	double tc2, mintogo, minutes, hours;
	

	/* initialize random numbers generator */
	initrandom();
	
	/* start */
	printf("\n*******************************************************");
	printf("**********************\n");
	printf("** PUZZLE %s:  Inferring maximum ", VERSION);
	printf("likelihood trees with quartet puzzling **\n");
	printf("*******************************************************");
	printf("**********************\n\n");

#ifdef SIMULATION
printf("\n\n!! COMPILED FOR SIMULATION STUDY (8 TAXA) !!\n\n");
#endif

#ifdef CONTBAYES
printf("\n\n!! COMPILED FOR QP USING CONINOUS WEIGHTS !!\n\n");
#endif

	/* get sequences */
	printf("\nINFO:  Looking for sequence input data\n");
	openfiletoread(&seqfp, INFILE);
	getsizesites(seqfp);
	getdataset(seqfp);
	closefile(seqfp);
	printf("\nINFO:  Contents: %d taxa with %d sites\n",
		Maxspc, Maxseqc);
	data_optn = guessdatatype();
	printf("\nINFO:  Data type: %s\n",
		(data_optn ? "Amino acids" : "Nucleotides"));
	
	/* translate characters into format used by ML engine */
	nuc_optn = TRUE; 
	SH_optn = FALSE;
	Seqchar = NULL;
	translatedataset();
	
	/* estimate base frequencies from data set */
	Freqtpm = NULL;
	Basecomp = NULL;
	estimatebasefreqs();
	
	/* estimate model of sequence evolution */
	printf("\nINFO:  Inferring model of sequence evolution\n");
	guessmodel();

	/* check for input tree file */
	if ((itfp = fopen(INTREE, "r")) != NULL) {
		fclose(itfp);
		printf("\nINFO:  Input tree file detected\n");
		puzzlemode = 1;
	} else {
		puzzlemode = 0;
	}

	/* set options interactively */
	printf("\nINFO:  Switching to recommended settings\n");
	setoptions();
	
	/* start main timer */
	time(&Starttime);

	/* symmetrize doublet frequencies if specified */
	symdoublets();

	/* initialise ML */
	printf("\nINFO:  Starting maximum likelihood engine\n");
	mlstart();

	/* show base frequencies that are actually used in the ML calculation */
	showfrequencies();

	/* estimate parameters from data set */
	if (lambda_optim || rate_optim || (data_optn == 0 && optim_optn)) {	
		printf("\nINFO:  Estimating missing parameters");
		printf(" from data set\n");
		optimparams();
		printf("\nINFO:  Estimation results:\n\n");
		if (data_optn == 0 && optim_optn) {
			printf("        Transition/transversion parameter:");
			printf(" %.2f +/- %.2f\n", AlphaBeta/2.0, tserr);
			if (TN_optn) {
				printf("        Y/R transition parameter:");
				printf(" %.2f +/- %.2f\n", AlphaYR, yrerr);
			}
		}
		if (rate_optim) {
			printf("        Rate heterogenity parameter:");
			printf(" %.2f +/- %.2f\n", Galpha, gaerr);
		}
		if (lambda_optim) {
			printf("        Rate correlation parameter:");
			printf(" %.2f +/- %.2f\n", lambda, laerr);
		}
	}
	
	/* show expected Ts/Tv ratio */
	if (data_optn == 0) shownucratios();
	
	/* show rate heterogenity categories */
	if (numcats > 1) showrates();
	
	/* show average rate patch length */
	if (numcats > 1) {
		printf("\nINFO:  ");
		showpatchlength(stdout);
		printf("\n");
	}

	/* computing ML distances */
	printf("\nINFO:  Computing maximum likelihood distances\n");
	computedistan();
	openfiletowrite(&dfp, DISTANCES);
	putdistance(dfp);
	closefile(dfp);
	printf("\nINFO:  Distances written to distance output file (%s)\n", DISTANCES);

	if (puzzlemode == 0) { /* Quartet Puzzling */

		/* allocate variables used for randomizing input order */
		trueID = new_ivector(Maxspc);
		freeID = new_ivector(Maxspc);
	
		/* allocate memory for quartets */
		printf("\nINFO:  Preparing for quartet information storage\n");
#ifdef CONTBAYES
		mallocquartetweights(Maxspc);
#else
		mallocquartets(Maxspc);
#endif
	
		/* prepare for consensus tree analysis */
		initconsensus();
		
		/* compute quartets */
		printf("\nINFO:  Computing quartet maximum likelihood trees\n");
		computeallquartets();
	
		printf(
			"\nINFO:  Performing %lu puzzling steps\n", Numtrial);
	
		/* start timer - percentage of completed trees */
		time(&time0);
		time1 = time0;
	
		for (Currtrial = 0; Currtrial < Numtrial; Currtrial++) {
			
			/* randomize input order */		
			for (k = 0; k < Maxspc; k++) freeID[k] = TRUE;
			for (k = 0; k < Maxspc; k++) {
				/* random integer in the range [0, Maxspc-1-k] */
				ID = randominteger(Maxspc - k);
				i = -1;
				j = -1;
				do {
					i++;
					if (freeID[i] == TRUE) j++; 
				} while ( j != ID);
				trueID[k] = i;
				freeID[i] = FALSE;
			}

			/* initialize tree */
			inittree();

			/* adding all other leafs */
			for (i = 3; i < Maxspc; i++) { 
					
				/* clear all edgeinfos or edgescores */
#ifdef CONTBAYES
				resetedgescore();
#else
				resetedgeinfo();
#endif
				
				/* clear counter of quartets */
				nqused = 0;

				/*
				 * core of quartet puzzling algorithm
				 */

				for (a = 0; a < nextleaf - 2; a++)
					for (b = a + 1; b < nextleaf - 1; b++)
						for (c = b + 1; c < nextleaf; c++) {

#ifdef CONTBAYES		
							/* assign a score equal to the weights for all
							   three possible topologies to each edge */
							assignscores(a, b, c, i);
#else
							/*	check which two _leaves_ out of a, b, c
							    are closer related to each other than
							    to leaf i according to a least squares
							    fit of the continous Baysian weights to the
							    seven trivial "attractive regions". We assign 
							    a score of 1 to all edges between these two leaves
							    chooseA and chooseB */

							checkquartet(a, b, c, i);
							incrementedgeinfo(chooseA, chooseB);
#endif	
								
							nqused++;
							
							/* generate message every 15 minutes */
								
							/* check timer */
							time(&time2);
							if ( (time2 - time1) > 900) {
								/* every 900 seconds */
								/* percentage of completed trees */
								tc2 = 100.0*Currtrial/Numtrial + 
										100.0*nqused/Numquartets/Numtrial;
								mintogo = (100.0-tc2) *
									(double) (time2-time0)/60.0/tc2;
								hours = floor(mintogo/60.0);
								minutes = mintogo - 60.0*hours;
								printf("\n        %2.2f%%", tc2);
								printf(" of the puzzling");
								printf(" completed (remaining");
								printf(" time: %.0f", hours);
								printf(" hours %.0f", minutes);
								printf(" minutes)\n");
								time1 = time2;
							}
						}

				/* find out which edge has the lowest edgeinfo */
#ifdef CONTBAYES
				minimumedgescore();
#else
				minimumedgeinfo();
#endif
				
				/* add the next leaf on minedge */
				addnextleaf(minedge);
			}
	
			/* compute bipartitions of current tree */
			computebiparts();
			makenewsplitentries();

			/* free tree before building the next tree */
			freetree();
		}

		/* garbage collection */
		free(splitcomp);
		free_ivector(freeID);
		free_ivector(trueID);
		free_cmatrix(biparts);
#ifdef CONTBAYES
		freequartetweights();
#else
		freequartets();
#endif

		/* compute majority rule consensus tree */
		printf("\nINFO:  Computing majority rule consensus tree\n");
		makeconsensus();
	
	}

	free_cmatrix(Seqchar);
	free_cmatrix(seqchars);
	
	/* write consensus tree */
	if (puzzlemode == 0) {
		openfiletowrite(&tfp,TREEFILE);
		writeconsensustree(tfp);
		closefile(tfp);
	}


#ifndef SIMULATION	
	/* compute ML branch lengths on QP tree */
	if (puzzlemode == 0) {
		printf(
			"\nINFO:  Computing maximum likelihood branch lengths\n");
		openfiletoread(&itfp, TREEFILE);
		usertree_lklhd(itfp);
		closefile(itfp);
		/* write ML branch length tree back to TREEFILE */
		openfiletowrite(&itfp, TREEFILE);
		fputphylogeny(itfp);
		closefile(itfp);
		printf(
			"\nINFO:  Resulting tree written to tree output file (%s)\n", TREEFILE);
	}
	
	if (puzzlemode == 1) {
		printf(
			"\nINFO:  Computing maximum likelihood branch lengths on user tree\n");
		openfiletoread(&itfp, INTREE);
		usertree_lklhd(itfp);
		closefile(itfp);
		/* write ML branch length tree to TREEFILE */
		openfiletowrite(&tfp, TREEFILE);
		fputphylogeny(tfp);
		closefile(tfp);
		printf(
			"\nINFO:  Resulting tree written to tree output file (%s)\n", TREEFILE);
	}
#endif

	/* stop timer (outfile needs runtime) */
	time(&Stoptime);

	/* write output file */
	openfiletowrite(&ofp, OUTFILE);
	writeoutputfile(ofp);
	closefile(ofp);

	printf(
		"\nINFO:  All information about this run written to output file (%s)\n", OUTFILE);
	mlfinish();

	/* runtime message */
	printf(
		"\nINFO:  This run took %lu seconds (= %.1f minutes = %.1f hours)\n\n",
			Stoptime - Starttime, (Stoptime - Starttime)/60.,
			(Stoptime - Starttime)/3600.);

	/* free memory */
	if (puzzlemode == 0) {
		free(splitfreqs);
		free(splitpatterns);
		free(splitsizes);
		free_ivector(consconfid);
		free_ivector(conssizes);
		free_cmatrix(consbiparts);
	}
	free_cmatrix(Identif);
	free_dvector(Freqtpm);
	free_imatrix(Basecomp);
}

/* compare function for ulint - sort largest numbers first */
int ulintcmp(const ulint *a, const ulint *b)
{
	if (*a > *b) return (-1);
	else if (*a < *b) return 1;
	else return 0;
}

/* compare function for double - sort largest numbers first */
int dcmp(const double *a, const double *b)
{
	if (*a > *b) return (-1);
	else if (*a < *b) return 1;
	else return 0;
}

/* compare function for int - sort smallest numbers first */
int intcmp(const int *a, const int *b)
{
	if (*a < *b) return (-1);
	else if (*a > *b) return 1;
	else return 0;
}
